import numpy as np

from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.utilities.physical_constants import me, mp

b_units = "code_magnetic"
e_units = "code_magnetic/c"
ra_units = "code_length / code_time**2"
rho_units = "code_mass / code_length**3"
vel_units = "code_velocity"

known_species_names = {
    "HI": "H_p0",
    "HII": "H_p1",
    "HeI": "He_p0",
    "HeII": "He_p1",
    "HeIII": "He_p2",
    "H2I": "H2_p0",
    "H2II": "H2_p1",
    "HM": "H_m1",
    "HeH": "HeH_p0",
    "DI": "D_p0",
    "DII": "D_p1",
    "HDI": "HD_p0",
    "Electron": "El",
    "OI": "O_p0",
    "OII": "O_p1",
    "OIII": "O_p2",
    "OIV": "O_p3",
    "OV": "O_p4",
    "OVI": "O_p5",
    "OVII": "O_p6",
    "OVIII": "O_p7",
    "OIX": "O_p8",
}

NODAL_FLAGS = {
    "BxF": [1, 0, 0],
    "ByF": [0, 1, 0],
    "BzF": [0, 0, 1],
    "Ex": [0, 1, 1],
    "Ey": [1, 0, 1],
    "Ez": [1, 1, 0],
    "AvgElec0": [0, 1, 1],
    "AvgElec1": [1, 0, 1],
    "AvgElec2": [1, 1, 0],
}


class EnzoFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("Cooling_Time", ("s", ["cooling_time"], None)),
        ("Dengo_Cooling_Rate", ("erg/g/s", [], None)),
        ("Grackle_Cooling_Rate", ("erg/s/cm**3", [], None)),
        ("HI_kph", ("1/code_time", ["H_p0_ionization_rate"], None)),
        ("HeI_kph", ("1/code_time", ["He_p0_ionization_rate"], None)),
        ("HeII_kph", ("1/code_time", ["He_p1_ionization_rate"], None)),
        ("H2I_kdiss", ("1/code_time", ["H2_p0_dissociation_rate"], None)),
        ("HM_kph", ("1/code_time", ["H_m1_ionization_rate"], None)),
        ("H2II_kdiss", ("1/code_time", ["H2_p1_dissociation_rate"], None)),
        ("Bx", (b_units, [], None)),
        ("By", (b_units, [], None)),
        ("Bz", (b_units, [], None)),
        ("BxF", (b_units, [], None)),
        ("ByF", (b_units, [], None)),
        ("BzF", (b_units, [], None)),
        ("Ex", (e_units, [], None)),
        ("Ey", (e_units, [], None)),
        ("Ez", (e_units, [], None)),
        ("AvgElec0", (e_units, [], None)),
        ("AvgElec1", (e_units, [], None)),
        ("AvgElec2", (e_units, [], None)),
        ("RadAccel1", (ra_units, ["radiation_acceleration_x"], None)),
        ("RadAccel2", (ra_units, ["radiation_acceleration_y"], None)),
        ("RadAccel3", (ra_units, ["radiation_acceleration_z"], None)),
        ("Dark_Matter_Density", (rho_units, ["dark_matter_density"], None)),
        ("Temperature", ("K", ["temperature"], None)),
        ("Dust_Temperature", ("K", ["dust_temperature"], None)),
        ("x-velocity", (vel_units, ["velocity_x"], None)),
        ("y-velocity", (vel_units, ["velocity_y"], None)),
        ("z-velocity", (vel_units, ["velocity_z"], None)),
        ("RaySegments", ("", ["ray_segments"], None)),
        ("PhotoGamma", ("eV/code_time", ["photo_gamma"], None)),
        ("PotentialField", ("code_velocity**2", ["gravitational_potential"], None)),
        ("Density", (rho_units, ["density"], None)),
        ("Metal_Density", (rho_units, ["metal_density"], None)),
        ("SN_Colour", (rho_units, [], None)),
        # Note: we do not alias Electron_Density to anything
        ("Electron_Density", (rho_units, [], None)),
    )

    known_particle_fields: KnownFieldsT = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("creation_time", ("code_time", [], None)),
        ("dynamical_time", ("code_time", [], None)),
        ("metallicity_fraction", ("code_metallicity", [], None)),
        ("metallicity", ("", [], None)),
        ("particle_type", ("", [], None)),
        ("particle_index", ("", [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("GridID", ("", [], None)),
        ("identifier", ("", ["particle_index"], None)),
        ("level", ("", [], None)),
        ("AccretionRate", ("code_mass/code_time", [], None)),
        ("AccretionRateTime", ("code_time", [], None)),
        ("AccretionRadius", ("code_length", [], None)),
        ("RadiationLifetime", ("code_time", [], None)),
    )

    def __init__(self, ds, field_list):
        hydro_method = ds.parameters.get("HydroMethod", None)
        if hydro_method is None:
            hydro_method = ds.parameters["Physics"]["Hydro"]["HydroMethod"]
        if hydro_method == 2:
            sl_left = slice(None, -2, None)
            sl_right = slice(1, -1, None)
            div_fac = 1.0
        else:
            sl_left = slice(None, -2, None)
            sl_right = slice(2, None, None)
            div_fac = 2.0
        slice_info = (sl_left, sl_right, div_fac)
        super().__init__(ds, field_list, slice_info)

        # setup nodal flag information
        for field in NODAL_FLAGS:
            if ("enzo", field) in self:
                finfo = self["enzo", field]
                finfo.nodal_flag = np.array(NODAL_FLAGS[field])

    def add_species_field(self, species):
        # This is currently specific to Enzo.  Hopefully in the future we will
        # have deeper integration with other systems, such as Dengo, to provide
        # better understanding of ionization and molecular states.
        #
        # We have several fields to add based on a given species field.  First
        # off, we add the species field itself.  Then we'll add a few more
        # items...
        #
        self.add_output_field(
            ("enzo", f"{species}_Density"),
            sampling_type="cell",
            take_log=True,
            units="code_mass/code_length**3",
        )
        yt_name = known_species_names[species]
        # don't alias electron density since mass is wrong
        if species != "Electron":
            self.alias(("gas", f"{yt_name}_density"), ("enzo", f"{species}_Density"))

    def setup_species_fields(self):
        species_names = [
            fn.rsplit("_Density")[0]
            for ft, fn in self.field_list
            if fn.endswith("_Density")
        ]
        species_names = [sp for sp in species_names if sp in known_species_names]

        def _electron_density(field, data):
            return data["enzo", "Electron_Density"] * (me / mp)

        self.add_field(
            ("gas", "El_density"),
            sampling_type="cell",
            function=_electron_density,
            units=self.ds.unit_system["density"],
        )
        for sp in species_names:
            self.add_species_field(sp)
            self.species_names.append(known_species_names[sp])
        self.species_names.sort()  # bb #1059

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        # Now we conditionally load a few other things.
        params = self.ds.parameters
        multi_species = params.get("MultiSpecies", None)
        dengo = params.get("DengoChemistryModel", 0)
        if multi_species is None:
            multi_species = params["Physics"]["AtomicPhysics"]["MultiSpecies"]
        if multi_species > 0 or dengo == 1:
            self.setup_species_fields()
        self.setup_energy_field()
        setup_magnetic_field_aliases(self, "enzo", [f"B{ax}" for ax in "xyz"])

    def setup_energy_field(self):
        unit_system = self.ds.unit_system
        # We check which type of field we need, and then we add it.
        ge_name = None
        te_name = None
        params = self.ds.parameters
        multi_species = params.get("MultiSpecies", None)
        if multi_species is None:
            multi_species = params["Physics"]["AtomicPhysics"]["MultiSpecies"]
        hydro_method = params.get("HydroMethod", None)
        if hydro_method is None:
            hydro_method = params["Physics"]["Hydro"]["HydroMethod"]
        dual_energy = params.get("DualEnergyFormalism", None)
        if dual_energy is None:
            dual_energy = params["Physics"]["Hydro"]["DualEnergyFormalism"]
        if ("enzo", "Gas_Energy") in self.field_list:
            ge_name = "Gas_Energy"
        elif ("enzo", "GasEnergy") in self.field_list:
            ge_name = "GasEnergy"
        if ("enzo", "Total_Energy") in self.field_list:
            te_name = "Total_Energy"
        elif ("enzo", "TotalEnergy") in self.field_list:
            te_name = "TotalEnergy"

        if hydro_method == 2 and te_name is not None:
            self.add_output_field(
                ("enzo", te_name), sampling_type="cell", units="code_velocity**2"
            )
            self.alias(("gas", "specific_thermal_energy"), ("enzo", te_name))

            def _ge_plus_kin(field, data):
                ret = data["enzo", te_name] + 0.5 * data["gas", "velocity_x"] ** 2.0
                if data.ds.dimensionality > 1:
                    ret += 0.5 * data["gas", "velocity_y"] ** 2.0
                if data.ds.dimensionality > 2:
                    ret += 0.5 * data["gas", "velocity_z"] ** 2.0
                return ret

            self.add_field(
                ("gas", "specific_total_energy"),
                sampling_type="cell",
                function=_ge_plus_kin,
                units=unit_system["specific_energy"],
            )
        elif dual_energy == 1:
            if te_name is not None:
                self.add_output_field(
                    ("enzo", te_name), sampling_type="cell", units="code_velocity**2"
                )
                self.alias(
                    ("gas", "specific_total_energy"),
                    ("enzo", te_name),
                    units=unit_system["specific_energy"],
                )
            if ge_name is not None:
                self.add_output_field(
                    ("enzo", ge_name), sampling_type="cell", units="code_velocity**2"
                )
                self.alias(
                    ("gas", "specific_thermal_energy"),
                    ("enzo", ge_name),
                    units=unit_system["specific_energy"],
                )
        elif hydro_method in (4, 6) and te_name is not None:
            self.add_output_field(
                ("enzo", te_name), sampling_type="cell", units="code_velocity**2"
            )

            # Subtract off B-field energy
            def _sub_b(field, data):
                ret = data["enzo", te_name] - 0.5 * data["gas", "velocity_x"] ** 2.0
                if data.ds.dimensionality > 1:
                    ret -= 0.5 * data["gas", "velocity_y"] ** 2.0
                if data.ds.dimensionality > 2:
                    ret -= 0.5 * data["gas", "velocity_z"] ** 2.0
                ret -= data["gas", "magnetic_energy_density"] / data["gas", "density"]
                return ret

            self.add_field(
                ("gas", "specific_thermal_energy"),
                sampling_type="cell",
                function=_sub_b,
                units=unit_system["specific_energy"],
            )
        elif te_name is not None:  # Otherwise, we assume TotalEnergy is kinetic+thermal
            self.add_output_field(
                ("enzo", te_name), sampling_type="cell", units="code_velocity**2"
            )
            self.alias(
                ("gas", "specific_total_energy"),
                ("enzo", te_name),
                units=unit_system["specific_energy"],
            )

            def _tot_minus_kin(field, data):
                ret = data["enzo", te_name] - 0.5 * data["gas", "velocity_x"] ** 2.0
                if data.ds.dimensionality > 1:
                    ret -= 0.5 * data["gas", "velocity_y"] ** 2.0
                if data.ds.dimensionality > 2:
                    ret -= 0.5 * data["gas", "velocity_z"] ** 2.0
                return ret

            self.add_field(
                ("gas", "specific_thermal_energy"),
                sampling_type="cell",
                function=_tot_minus_kin,
                units=unit_system["specific_energy"],
            )
        if multi_species == 0 and "Mu" in params:

            def _mean_molecular_weight(field, data):
                return params["Mu"] * data["index", "ones"]

            self.add_field(
                ("gas", "mean_molecular_weight"),
                sampling_type="cell",
                function=_mean_molecular_weight,
                units="",
            )

            def _number_density(field, data):
                return data["gas", "density"] / (mp * params["Mu"])

            self.add_field(
                ("gas", "number_density"),
                sampling_type="cell",
                function=_number_density,
                units=unit_system["number_density"],
            )

    def setup_particle_fields(self, ptype):
        def _age(field, data):
            return data.ds.current_time - data["all", "creation_time"]

        self.add_field(
            (ptype, "age"), sampling_type="particle", function=_age, units="yr"
        )

        super().setup_particle_fields(ptype)



"""
Title: conftest.py
Purpose: Contains fixtures for loading data.
"""

# Test data
sedov = "sedov/sedov_tst_0004.h5"


# Test parameters. Format:
# {test1: {param1 : [(val1, val2,...), (id1, id2,...)], param2 : ...}, test2: ...}
test_params = {
    "test_sedov_tunnel": {
        "axis": [(0, 1, 2), ("0", "1", "2")],
        "dobj": [(None, ("sphere", ("max", (0.1, "unitary")))), ("None", "sphere")],
        "weight": [(None, ("gas", "density")), ("None", "density")],
        "field": [(("gas", "density"), ("gas", "velocity_x")), ("density", "vx")],
    }
}


def pytest_generate_tests(metafunc):
    # Loop over each test in test_params
    for test_name, params in test_params.items():
        if metafunc.function.__name__ == test_name:
            # Parametrize
            for param_name, param_vals in params.items():
                metafunc.parametrize(param_name, param_vals[0], ids=param_vals[1])



import re
from typing import TypeAlias

import numpy as np

from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.units import YTQuantity
from yt.utilities.physical_constants import amu_cgs, boltzmann_constant_cgs, c

rho_units = "code_mass / code_length**3"
mom_units = "code_mass / (code_time * code_length**2)"
eden_units = "code_mass / (code_time**2 * code_length)"  # erg / cm^3


def _thermal_energy_density(field, data):
    # What we've got here is UEINT:
    # u here is velocity
    # E is energy density from the file
    #   rho e = rho E - rho * u * u / 2
    ke = (
        0.5
        * (
            data["gas", "momentum_density_x"] ** 2
            + data["gas", "momentum_density_y"] ** 2
            + data["gas", "momentum_density_z"] ** 2
        )
        / data["gas", "density"]
    )
    return data["boxlib", "eden"] - ke


def _specific_thermal_energy(field, data):
    # This is little e, so we take thermal_energy_density and divide by density
    return data["gas", "thermal_energy_density"] / data["gas", "density"]


def _temperature(field, data):
    mu = data.ds.parameters["mu"]
    gamma = data.ds.parameters["gamma"]
    tr = data["gas", "thermal_energy_density"] / data["gas", "density"]
    tr *= mu * amu_cgs / boltzmann_constant_cgs
    tr *= gamma - 1.0
    return tr


class WarpXFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("Bx", ("T", ["magnetic_field_x", "B_x"], None)),
        ("By", ("T", ["magnetic_field_y", "B_y"], None)),
        ("Bz", ("T", ["magnetic_field_z", "B_z"], None)),
        ("Ex", ("V/m", ["electric_field_x", "E_x"], None)),
        ("Ey", ("V/m", ["electric_field_y", "E_y"], None)),
        ("Ez", ("V/m", ["electric_field_z", "E_z"], None)),
        ("jx", ("A", ["current_x", "Jx", "J_x"], None)),
        ("jy", ("A", ["current_y", "Jy", "J_y"], None)),
        ("jz", ("A", ["current_z", "Jz", "J_z"], None)),
    )

    known_particle_fields: KnownFieldsT = (
        ("particle_weight", ("", ["particle_weighting"], None)),
        ("particle_position_x", ("m", [], None)),
        ("particle_position_y", ("m", [], None)),
        ("particle_position_z", ("m", [], None)),
        ("particle_velocity_x", ("m/s", [], None)),
        ("particle_velocity_y", ("m/s", [], None)),
        ("particle_velocity_z", ("m/s", [], None)),
        ("particle_momentum_x", ("kg*m/s", [], None)),
        ("particle_momentum_y", ("kg*m/s", [], None)),
        ("particle_momentum_z", ("kg*m/s", [], None)),
    )

    extra_union_fields = (
        ("kg", "particle_mass"),
        ("C", "particle_charge"),
        ("", "particle_ones"),
    )

    def __init__(self, ds, field_list):
        super().__init__(ds, field_list)

        # setup nodal flag information
        for field in ds.index.raw_fields:
            finfo = self.__getitem__(("raw", field))
            finfo.nodal_flag = ds.nodal_flags[field]

    def setup_fluid_fields(self):
        for field in self.known_other_fields:
            fname = field[0]
            self.alias(("mesh", fname), ("boxlib", fname))

    def setup_fluid_aliases(self):
        super().setup_fluid_aliases("mesh")

    def setup_particle_fields(self, ptype):
        def get_mass(field, data):
            species_mass = data.ds.index.parameters[ptype + "_mass"]
            return data[ptype, "particle_weight"] * YTQuantity(species_mass, "kg")

        self.add_field(
            (ptype, "particle_mass"),
            sampling_type="particle",
            function=get_mass,
            units="kg",
        )

        def get_charge(field, data):
            species_charge = data.ds.index.parameters[ptype + "_charge"]
            return data[ptype, "particle_weight"] * YTQuantity(species_charge, "C")

        self.add_field(
            (ptype, "particle_charge"),
            sampling_type="particle",
            function=get_charge,
            units="C",
        )

        def get_energy(field, data):
            p2 = (
                data[ptype, "particle_momentum_x"] ** 2
                + data[ptype, "particle_momentum_y"] ** 2
                + data[ptype, "particle_momentum_z"] ** 2
            )
            return np.sqrt(p2 * c**2 + data[ptype, "particle_mass"] ** 2 * c**4)

        self.add_field(
            (ptype, "particle_energy"),
            sampling_type="particle",
            function=get_energy,
            units="J",
        )

        def get_velocity_x(field, data):
            return (
                c**2
                * data[ptype, "particle_momentum_x"]
                / data[ptype, "particle_energy"]
            )

        def get_velocity_y(field, data):
            return (
                c**2
                * data[ptype, "particle_momentum_y"]
                / data[ptype, "particle_energy"]
            )

        def get_velocity_z(field, data):
            return (
                c**2
                * data[ptype, "particle_momentum_z"]
                / data[ptype, "particle_energy"]
            )

        self.add_field(
            (ptype, "particle_velocity_x"),
            sampling_type="particle",
            function=get_velocity_x,
            units="m/s",
        )

        self.add_field(
            (ptype, "particle_velocity_y"),
            sampling_type="particle",
            function=get_velocity_y,
            units="m/s",
        )

        self.add_field(
            (ptype, "particle_velocity_z"),
            sampling_type="particle",
            function=get_velocity_z,
            units="m/s",
        )

        super().setup_particle_fields(ptype)


class NyxFieldInfo(FieldInfoContainer):
    known_particle_fields: KnownFieldsT = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
    )


class BoxlibFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("density", (rho_units, ["density"], None)),
        ("eden", (eden_units, ["total_energy_density"], None)),
        ("xmom", (mom_units, ["momentum_density_x"], None)),
        ("ymom", (mom_units, ["momentum_density_y"], None)),
        ("zmom", (mom_units, ["momentum_density_z"], None)),
        ("temperature", ("K", ["temperature"], None)),
        ("Temp", ("K", ["temperature"], None)),
        ("x_velocity", ("cm/s", ["velocity_x"], None)),
        ("y_velocity", ("cm/s", ["velocity_y"], None)),
        ("z_velocity", ("cm/s", ["velocity_z"], None)),
        ("xvel", ("cm/s", ["velocity_x"], None)),
        ("yvel", ("cm/s", ["velocity_y"], None)),
        ("zvel", ("cm/s", ["velocity_z"], None)),
    )

    known_particle_fields: KnownFieldsT = (
        ("particle_mass", ("code_mass", [], None)),
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_momentum_x", ("code_mass*code_length/code_time", [], None)),
        ("particle_momentum_y", ("code_mass*code_length/code_time", [], None)),
        ("particle_momentum_z", ("code_mass*code_length/code_time", [], None)),
        # Note that these are *internal* agmomen
        ("particle_angmomen_x", ("code_length**2/code_time", [], None)),
        ("particle_angmomen_y", ("code_length**2/code_time", [], None)),
        ("particle_angmomen_z", ("code_length**2/code_time", [], None)),
        ("particle_id", ("", ["particle_index"], None)),
        ("particle_mdot", ("code_mass/code_time", [], None)),
        # "mlast",
        # "r",
        # "mdeut",
        # "n",
        # "burnstate",
        # "luminosity",
    )

    def setup_particle_fields(self, ptype):
        def _get_vel(axis):
            def velocity(field, data):
                return (
                    data[ptype, f"particle_momentum_{axis}"]
                    / data[ptype, "particle_mass"]
                )

            return velocity

        for ax in "xyz":
            self.add_field(
                (ptype, f"particle_velocity_{ax}"),
                sampling_type="particle",
                function=_get_vel(ax),
                units="code_length/code_time",
            )

        super().setup_particle_fields(ptype)

    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system
        # Now, let's figure out what fields are included.
        if any(f[1] == "xmom" for f in self.field_list):
            self.setup_momentum_to_velocity()
        elif any(f[1] == "xvel" for f in self.field_list):
            self.setup_velocity_to_momentum()
        self.add_field(
            ("gas", "specific_thermal_energy"),
            sampling_type="cell",
            function=_specific_thermal_energy,
            units=unit_system["specific_energy"],
        )
        self.add_field(
            ("gas", "thermal_energy_density"),
            sampling_type="cell",
            function=_thermal_energy_density,
            units=unit_system["pressure"],
        )
        if ("gas", "temperature") not in self.field_aliases:
            self.add_field(
                ("gas", "temperature"),
                sampling_type="cell",
                function=_temperature,
                units=unit_system["temperature"],
            )

    def setup_momentum_to_velocity(self):
        def _get_vel(axis):
            def velocity(field, data):
                return data["boxlib", f"{axis}mom"] / data["boxlib", "density"]

            return velocity

        for ax in "xyz":
            self.add_field(
                ("gas", f"velocity_{ax}"),
                sampling_type="cell",
                function=_get_vel(ax),
                units=self.ds.unit_system["velocity"],
            )

    def setup_velocity_to_momentum(self):
        def _get_mom(axis):
            def momentum(field, data):
                return data["boxlib", f"{axis}vel"] * data["boxlib", "density"]

            return momentum

        for ax in "xyz":
            self.add_field(
                ("gas", f"momentum_density_{ax}"),
                sampling_type="cell",
                function=_get_mom(ax),
                units=mom_units,
            )


class CastroFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("density", ("g/cm**3", ["density"], r"\rho")),
        ("xmom", ("g/(cm**2 * s)", ["momentum_density_x"], r"\rho u")),
        ("ymom", ("g/(cm**2 * s)", ["momentum_density_y"], r"\rho v")),
        ("zmom", ("g/(cm**2 * s)", ["momentum_density_z"], r"\rho w")),
        # velocity components are not always present
        ("x_velocity", ("cm/s", ["velocity_x"], r"u")),
        ("y_velocity", ("cm/s", ["velocity_y"], r"v")),
        ("z_velocity", ("cm/s", ["velocity_z"], r"w")),
        ("rho_E", ("erg/cm**3", ["total_energy_density"], r"\rho E")),
        # internal energy density (not just thermal)
        ("rho_e", ("erg/cm**3", [], r"\rho e")),
        ("Temp", ("K", ["temperature"], r"T")),
        ("grav_x", ("cm/s**2", [], r"\mathbf{g} \cdot \mathbf{e}_x")),
        ("grav_y", ("cm/s**2", [], r"\mathbf{g} \cdot \mathbf{e}_y")),
        ("grav_z", ("cm/s**2", [], r"\mathbf{g} \cdot \mathbf{e}_z")),
        ("pressure", ("dyne/cm**2", [], r"p")),
        (
            "kineng",
            ("erg/cm**3", ["kinetic_energy_density"], r"\frac{1}{2}\rho|\mathbf{U}|^2"),
        ),
        ("soundspeed", ("cm/s", ["sound_speed"], "Sound Speed")),
        ("MachNumber", ("", ["mach_number"], "Mach Number")),
        ("abar", ("", [], r"$\bar{A}$")),
        ("Ye", ("", [], r"$Y_e$")),
        ("entropy", ("erg/(g*K)", ["entropy"], r"s")),
        ("magvort", ("1/s", ["vorticity_magnitude"], r"|\nabla \times \mathbf{U}|")),
        ("divu", ("1/s", ["velocity_divergence"], r"\nabla \cdot \mathbf{U}")),
        ("eint_E", ("erg/g", [], r"e(E,U)")),
        ("eint_e", ("erg/g", [], r"e")),
        ("magvel", ("cm/s", ["velocity_magnitude"], r"|\mathbf{U}|")),
        ("radvel", ("cm/s", ["radial_velocity"], r"\mathbf{U} \cdot \mathbf{e}_r")),
        ("magmom", ("g*cm/s", ["momentum_magnitude"], r"\rho |\mathbf{U}|")),
        ("maggrav", ("cm/s**2", [], r"|\mathbf{g}|")),
        ("phiGrav", ("erg/g", [], r"\Phi")),
        ("enuc", ("erg/(g*s)", [], r"\dot{e}_{\rm{nuc}}")),
        ("rho_enuc", ("erg/(cm**3*s)", [], r"\rho \dot{e}_{\rm{nuc}}")),
        ("angular_momentum_x", ("g/(cm*s)", [], r"\ell_x")),
        ("angular_momentum_y", ("g/(cm*s)", [], r"\ell_y")),
        ("angular_momentum_z", ("g/(cm*s)", [], r"\ell_z")),
        ("phiRot", ("erg/g", [], r"\Phi_{\rm{rot}}")),
        ("rot_x", ("cm/s**2", [], r"\mathbf{f}_{\rm{rot}} \cdot \mathbf{e}_x")),
        ("rot_y", ("cm/s**2", [], r"\mathbf{f}_{\rm{rot}} \cdot \mathbf{e}_y")),
        ("rot_z", ("cm/s**2", [], r"\mathbf{f}_{\rm{rot}} \cdot \mathbf{e}_z")),
    )

    known_particle_fields: KnownFieldsT = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
    )

    def setup_fluid_fields(self):
        # add X's
        for _, field in self.ds.field_list:
            if field.startswith("X("):
                # We have a fraction
                sub = Substance(field)
                # Overwrite field to use nicer tex_label display_name
                self.add_output_field(
                    ("boxlib", field),
                    sampling_type="cell",
                    units="",
                    display_name=rf"X\left({sub.to_tex()}\right)",
                )
                self.alias(("gas", f"{sub}_fraction"), ("boxlib", field), units="")
                func = _create_density_func(("gas", f"{sub}_fraction"))
                self.add_field(
                    name=("gas", f"{sub}_density"),
                    sampling_type="cell",
                    function=func,
                    units=self.ds.unit_system["density"],
                    display_name=rf"\rho {sub.to_tex()}",
                )


class MaestroFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("density", ("g/cm**3", ["density"], None)),
        ("x_vel", ("cm/s", ["velocity_x"], r"\tilde{u}")),
        ("y_vel", ("cm/s", ["velocity_y"], r"\tilde{v}")),
        ("z_vel", ("cm/s", ["velocity_z"], r"\tilde{w}")),
        (
            "magvel",
            (
                "cm/s",
                ["velocity_magnitude"],
                r"|\tilde{\mathbf{U}} + w_0 \mathbf{e}_r|",
            ),
        ),
        (
            "radial_velocity",
            ("cm/s", ["radial_velocity"], r"\mathbf{U}\cdot \mathbf{e}_r"),
        ),
        ("circum_velocity", ("cm/s", ["tangential_velocity"], r"U - U\cdot e_r")),
        ("tfromp", ("K", [], "T(\\rho,p,X)")),
        ("tfromh", ("K", [], "T(\\rho,h,X)")),
        ("Machnumber", ("", ["mach_number"], "M")),
        ("S", ("1/s", [], None)),
        ("ad_excess", ("", [], r"\nabla - \nabla_\mathrm{ad}")),
        ("deltaT", ("", [], "[T(\\rho,h,X) - T(\\rho,p,X)]/T(\\rho,h,X)")),
        ("deltagamma", ("", [], r"\Gamma_1 - \overline{\Gamma_1}")),
        ("deltap", ("", [], "[p(\\rho,h,X) - p_0] / p_0")),
        ("divw0", ("1/s", [], r"\nabla \cdot \mathbf{w}_0")),
        # Specific entropy
        ("entropy", ("erg/(g*K)", ["entropy"], "s")),
        ("entropypert", ("", [], r"[s - \overline{s}] / \overline{s}")),
        ("enucdot", ("erg/(g*s)", [], r"\dot{\epsilon}_{nuc}")),
        ("Hext", ("erg/(g*s)", [], "H_{ext}")),
        # Perturbational pressure grad
        ("gpi_x", ("dyne/cm**3", [], r"\left(\nabla\pi\right)_x")),
        ("gpi_y", ("dyne/cm**3", [], r"\left(\nabla\pi\right)_y")),
        ("gpi_z", ("dyne/cm**3", [], r"\left(\nabla\pi\right)_z")),
        ("h", ("erg/g", [], "h")),
        ("h0", ("erg/g", [], "h_0")),
        # Momentum cannot be computed because we need to include base and
        # full state.
        ("momentum", ("g*cm/s", ["momentum_magnitude"], r"\rho |\mathbf{U}|")),
        ("p0", ("erg/cm**3", [], "p_0")),
        ("p0pluspi", ("erg/cm**3", [], r"p_0 + \pi")),
        ("pi", ("erg/cm**3", [], r"\pi")),
        ("pioverp0", ("", [], r"\pi/p_0")),
        # Base state density
        ("rho0", ("g/cm**3", [], "\\rho_0")),
        ("rhoh", ("erg/cm**3", ["enthalpy_density"], "(\\rho h)")),
        # Base state enthalpy density
        ("rhoh0", ("erg/cm**3", [], "(\\rho h)_0")),
        ("rhohpert", ("erg/cm**3", [], "(\\rho h)^\\prime")),
        ("rhopert", ("g/cm**3", [], "\\rho^\\prime")),
        ("soundspeed", ("cm/s", ["sound_speed"], None)),
        ("sponge", ("", [], None)),
        ("tpert", ("K", [], r"T - \overline{T}")),
        # Again, base state -- so we can't compute ourselves.
        ("vort", ("1/s", ["vorticity_magnitude"], r"|\nabla\times\tilde{U}|")),
        # Base state
        ("w0_x", ("cm/s", [], "(w_0)_x")),
        ("w0_y", ("cm/s", [], "(w_0)_y")),
        ("w0_z", ("cm/s", [], "(w_0)_z")),
    )

    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system
        # pick the correct temperature field
        tfromp = False
        if "use_tfromp" in self.ds.parameters:
            # original MAESTRO (F90) code
            tfromp = self.ds.parameters["use_tfromp"]
        elif "maestro.use_tfromp" in self.ds.parameters:
            # new MAESTROeX (C++) code
            tfromp = self.ds.parameters["maestro.use_tfromp"]

        if tfromp:
            self.alias(
                ("gas", "temperature"),
                ("boxlib", "tfromp"),
                units=unit_system["temperature"],
            )
        else:
            self.alias(
                ("gas", "temperature"),
                ("boxlib", "tfromh"),
                units=unit_system["temperature"],
            )

        # Add X's and omegadots, units of 1/s
        for _, field in self.ds.field_list:
            if field.startswith("X("):
                # We have a mass fraction
                sub = Substance(field)
                # Overwrite field to use nicer tex_label display_name
                self.add_output_field(
                    ("boxlib", field),
                    sampling_type="cell",
                    units="",
                    display_name=rf"X\left({sub.to_tex()}\right)",
                )
                self.alias(("gas", f"{sub}_fraction"), ("boxlib", field), units="")
                func = _create_density_func(("gas", f"{sub}_fraction"))
                self.add_field(
                    name=("gas", f"{sub}_density"),
                    sampling_type="cell",
                    function=func,
                    units=unit_system["density"],
                    display_name=rf"\rho {sub.to_tex()}",
                )

            elif field.startswith("omegadot("):
                sub = Substance(field)
                display_name = rf"\dot{{\omega}}\left[{sub.to_tex()}\right]"
                # Overwrite field to use nicer tex_label'ed display_name
                self.add_output_field(
                    ("boxlib", field),
                    sampling_type="cell",
                    units=unit_system["frequency"],
                    display_name=display_name,
                )
                self.alias(
                    ("gas", f"{sub}_creation_rate"),
                    ("boxlib", field),
                    units=unit_system["frequency"],
                )


substance_expr_re = re.compile(r"\(([a-zA-Z][a-zA-Z0-9]*)\)")
substance_elements_re = re.compile(r"(?P<element>[a-zA-Z]+)(?P<digits>\d*)")
SubstanceSpec: TypeAlias = list[tuple[str, int]]


class Substance:
    def __init__(self, data: str) -> None:
        if (m := substance_expr_re.search(data)) is None:
            raise ValueError(f"{data!r} doesn't match expected regular expression")
        sub_str = m.group()
        constituents = substance_elements_re.findall(sub_str)

        # 0 is used as a sentinel value to mark descriptive names
        default_value = 1 if len(constituents) > 1 else 0
        self._spec: SubstanceSpec = [
            (name, int(count or default_value)) for (name, count) in constituents
        ]

    def get_spec(self) -> SubstanceSpec:
        return self._spec.copy()

    def is_isotope(self) -> bool:
        return len(self._spec) == 1 and self._spec[0][1] > 0

    def is_molecule(self) -> bool:
        return len(self._spec) != 1

    def is_descriptive_name(self) -> bool:
        return len(self._spec) == 1 and self._spec[0][1] == 0

    def __str__(self) -> str:
        return "".join(
            f"{element}{count if count > 1 else ''}" for element, count in self._spec
        )

    def _to_tex_isotope(self) -> str:
        element, count = self._spec[0]
        return rf"^{{{count}}}{element}"

    def _to_tex_molecule(self) -> str:
        return "".join(
            rf"{element}_{{{count if count>1 else ''}}}"
            for element, count in self._spec
        )

    def _to_tex_descriptive(self) -> str:
        return str(self)

    def to_tex(self) -> str:
        if self.is_isotope():
            return self._to_tex_isotope()
        elif self.is_molecule():
            return self._to_tex_molecule()
        elif self.is_descriptive_name():
            return self._to_tex_descriptive()
        else:
            # should only be reachable in case of a regular expression defect
            raise RuntimeError


def _create_density_func(field_name):
    def _func(field, data):
        return data[field_name] * data["gas", "density"]

    return _func



cdef class SphereSelector(SelectorObject):
    cdef public np.float64_t radius
    cdef public np.float64_t radius2
    cdef public np.float64_t center[3]
    cdef np.float64_t bbox[3][2]
    cdef public bint check_box[3]

    def __init__(self, dobj):
        for i in range(3):
            self.center[i] = _ensure_code(dobj.center[i])
        self.radius = _ensure_code(dobj.radius)
        self.radius2 = self.radius * self.radius
        self.set_bbox(_ensure_code(dobj.center))
        for i in range(3):
            if self.bbox[i][0] < dobj.ds.domain_left_edge[i]:
                self.check_box[i] = False
            elif self.bbox[i][1] > dobj.ds.domain_right_edge[i]:
                self.check_box[i] = False
            else:
                self.check_box[i] = True

    def set_bbox(self, center):
        for i in range(3):
            self.center[i] = center[i]
            self.bbox[i][0] = self.center[i] - self.radius
            self.bbox[i][1] = self.center[i] + self.radius

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        # sphere center either inside cell or center of cell lies inside sphere
        if (pos[0] - 0.5*dds[0] <= self.center[0] <= pos[0]+0.5*dds[0] and
            pos[1] - 0.5*dds[1] <= self.center[1] <= pos[1]+0.5*dds[1] and
            pos[2] - 0.5*dds[2] <= self.center[2] <= pos[2]+0.5*dds[2]):
            return 1
        return self.select_point(pos)
        # # langmm: added to allow sphere to intersect edge/corner of cell
        # cdef np.float64_t LE[3]
        # cdef np.float64_t RE[3]
        # cdef int i
        # for i in range(3):
        #     LE[i] = pos[i] - 0.5*dds[i]
        #     RE[i] = pos[i] + 0.5*dds[i]
        # return self.select_bbox(LE, RE)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        cdef int i
        cdef np.float64_t dist, dist2 = 0
        for i in range(3):
            if self.check_box[i] and \
              (pos[i] < self.bbox[i][0] or
               pos[i] > self.bbox[i][1]):
                return 0
            dist = _periodic_dist(pos[i], self.center[i], self.domain_width[i],
                                  self.periodicity[i])
            dist2 += dist*dist
            if dist2 > self.radius2: return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        cdef int i
        cdef np.float64_t dist, dist2 = 0
        for i in range(3):
            dist = self.periodic_difference(pos[i], self.center[i], i)
            dist2 += dist*dist
        dist = self.radius+radius
        if dist2 <= dist*dist: return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        cdef np.float64_t box_center, relcenter, closest, dist, edge
        cdef int i
        if (left_edge[0] <= self.center[0] < right_edge[0] and
            left_edge[1] <= self.center[1] < right_edge[1] and
            left_edge[2] <= self.center[2] < right_edge[2]):
            return 1
        for i in range(3):
            if not self.check_box[i]: continue
            if right_edge[i] < self.bbox[i][0] or \
               left_edge[i] > self.bbox[i][1]:
                return 0
        # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
        dist = 0
        for i in range(3):
            # Early terminate
            box_center = (right_edge[i] + left_edge[i])/2.0
            relcenter = self.periodic_difference(box_center, self.center[i], i)
            edge = right_edge[i] - left_edge[i]
            closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
            dist += closest*closest
            if dist > self.radius2: return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        cdef np.float64_t box_center, relcenter, closest, farthest, cdist, fdist, edge
        cdef int i
        if (left_edge[0] <= self.center[0] <= right_edge[0] and
            left_edge[1] <= self.center[1] <= right_edge[1] and
            left_edge[2] <= self.center[2] <= right_edge[2]):
            fdist = 0
            for i in range(3):
                edge = right_edge[i] - left_edge[i]
                box_center = (right_edge[i] + left_edge[i])/2.0
                relcenter = self.periodic_difference(
                    box_center, self.center[i], i)
                if relcenter >= 0:
                    farthest = relcenter + edge/2.0
                else:
                    farthest = relcenter - edge/2.0
                # farthest = relcenter + fclip(relcenter, -edge/2.0, edge/2.0)
                fdist += farthest*farthest
                if fdist >= self.radius2:
                    return 2  # Box extends outside sphere
            return 1  # Box entirely inside sphere
        for i in range(3):
            if not self.check_box[i]: continue
            if right_edge[i] < self.bbox[i][0] or \
               left_edge[i] > self.bbox[i][1]:
                return 0  # Box outside sphere bounding box
        # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
        cdist = 0
        fdist = 0
        for i in range(3):
            # Early terminate
            box_center = (right_edge[i] + left_edge[i])/2.0
            relcenter = self.periodic_difference(box_center, self.center[i], i)
            edge = right_edge[i] - left_edge[i]
            closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
            if relcenter >= 0:
                farthest = relcenter + edge/2.0
            else:
                farthest = relcenter - edge/2.0
            #farthest = relcenter + fclip(relcenter, -edge/2.0, edge/2.0)
            cdist += closest*closest
            fdist += farthest*farthest
            if cdist > self.radius2:
                return 0  # Box does not overlap sphere
        if fdist < self.radius2:
            return 1  # Sphere extends to entirely contain box
        else:
            return 2  # Sphere only partially overlaps box

    def _hash_vals(self):
        return (("radius", self.radius),
                ("radius2", self.radius2),
                ("center[0]", self.center[0]),
                ("center[1]", self.center[1]),
                ("center[2]", self.center[2]))

    def _get_state_attnames(self):
        return ("radius", "radius2", "center", "check_box")

    def __setstate__(self, hashes):
        super(SphereSelector, self).__setstate__(hashes)
        self.set_bbox(self.center)


sphere_selector = SphereSelector



# distutils: libraries = STD_LIBS
"""
Bit array functions



"""


import numpy as np

cimport cython
cimport numpy as np


cdef class bitarray:

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def __cinit__(self, np.int64_t size = -1,
                  np.ndarray[np.uint8_t, ndim=1, cast=True] arr = None):
        r"""This is a bitarray, which flips individual bits to on/off inside a
        uint8 container array.

        By encoding on/off inside each bit in a uint8 array, we can compress
        boolean information down by up to a factor of 8.  Either an input array
        or a size must be provided.

        Parameters
        ----------
        size : int
            The size we should pre-allocate.
        arr : array-like
            An input array to turn into a bitarray.

        Examples
        --------

        >>> arr_in1 = np.array([True, True, False])
        >>> arr_in2 = np.array([False, True, True])
        >>> a = ba.bitarray(arr = arr_in1)
        >>> b = ba.bitarray(arr = arr_in2)
        >>> print(a & b)
        >>> print (a & b).as_bool_array()

        """
        cdef np.uint64_t i
        if size == -1 and arr is None:
            raise RuntimeError
        elif size == -1:
            size = arr.size
        elif size != -1 and arr is not None:
            if size != arr.size:
                raise RuntimeError
        self.buf_size = (size >> 3)
        cdef np.uint8_t bitmask = 255
        if (size & 7) != 0:
            # We need an extra one if we've got any lingering bits
            self.buf_size += 1
            bitmask = 0
            for i in range(size & 7):
                bitmask |= (1<<i)
        self.final_bitmask = bitmask
        cdef np.ndarray[np.uint8_t] ibuf_t
        ibuf_t = self.ibuf = np.zeros(self.buf_size, "uint8")
        self.buf = <np.uint8_t *> ibuf_t.data
        self.size = size
        if arr is not None:
            self.set_from_array(arr)
        else:
            for i in range(self.buf_size):
                self.buf[i] = 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def set_from_array(self, np.ndarray[np.uint8_t, cast=True] arr not None):
        r"""Given an array that is either uint8_t or boolean, set the values of
        this array to match it.

        Parameters
        ----------
        arr : array, castable to uint8
            The array we set from.
        """
        cdef np.uint64_t i, j
        cdef np.uint8_t *btemp = self.buf
        arr = np.ascontiguousarray(arr)
        j = 0
        for i in range(self.size):
            btemp[i >> 3] = btemp[i >> 3] | (arr[i] << (j))
            j += 1
            if j == 8:
                j = 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def as_bool_array(self):
        r"""Return a copy of this array, as a boolean array.

        All of the values encoded in this bitarray are expanded into boolean
        values in a new array and returned.

        Returns
        -------
        arr : numpy array of type bool
            The uint8 values expanded into boolean values

        """
        cdef np.uint64_t i, j
        cdef np.uint8_t *btemp = self.buf
        cdef np.ndarray[np.uint8_t, ndim=1] output
        output = np.zeros(self.size, "uint8")
        j = 0
        for i in range(self.size):
            output[i] = (btemp[i >> 3] >> (j)) & 1
            j += 1
            if j == 8:
                j = 0
        return output.astype("bool")

    cdef void _set_value(self, np.uint64_t ind, np.uint8_t val):
        ba_set_value(self.buf, ind, val)

    def set_value(self, np.uint64_t ind, np.uint8_t val):
        r"""Set the on/off value of a given bit.

        Modify the value encoded in a given index.

        Parameters
        ----------
        ind : int
            The index to query in the bitarray.
        val : bool or uint8_t
            What to set the index to

        Examples
        --------

        >>> arr_in = np.array([True, True, False])
        >>> a = ba.bitarray(arr = arr_in)
        >>> print(a.set_value(2, 1))

        """
        ba_set_value(self.buf, ind, val)

    cdef np.uint8_t _query_value(self, np.uint64_t ind):
        return ba_get_value(self.buf, ind)

    def query_value(self, np.uint64_t ind):
        r"""Query the on/off value of a given bit.

        Return the value encoded in a given index.

        Parameters
        ----------
        ind : int
            The index to query in the bitarray.

        Examples
        --------

        >>> arr_in = np.array([True, True, False])
        >>> a = ba.bitarray(arr = arr_in)
        >>> print(a.query_value(2))

        """
        return ba_get_value(self.buf, ind)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _set_range(self, np.uint64_t start, np.uint64_t stop, np.uint8_t val):
        ba_set_range(self.buf, start, stop, val)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def set_range(self, np.uint64_t start, np.uint64_t stop, np.uint8_t val):
        r"""Set a range of values to on/off.  Uses slice-style indexing.

        No return value.

        Parameters
        ----------
        start : int
            The starting component of a slice.
        stop : int
            The ending component of a slice.
        val : bool or uint8_t
            What to set the range to

        Examples
        --------

        >>> arr_in = np.array([True, True, False, True, True, False])
        >>> a = ba.bitarray(arr = arr_in)
        >>> a.set_range(0, 3, 0)

        """
        ba_set_range(self.buf, start, stop, val)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef np.uint64_t _count(self):
        cdef np.uint64_t count = 0
        cdef np.uint64_t i
        self.buf[self.buf_size - 1] &= self.final_bitmask
        for i in range(self.buf_size):
            count += _num_set_bits(self.buf[i])
        return count


    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def count(self):
        r"""Count the number of values set in the array.

        Parameters
        ----------

        Examples
        --------

        >>> arr_in = np.array([True, True, False, True, True, False])
        >>> a = ba.bitarray(arr = arr_in)
        >>> a.count()

        """
        return self._count()

    cdef bitarray _logical_and(self, bitarray other, bitarray result = None):
        # Create a place to put it.  Note that we might have trailing values,
        # we actually need to reset the ending set.
        if other.size != self.size:
            raise IndexError
        if result is None:
            result = bitarray(self.size)
        for i in range(self.buf_size):
            result.buf[i] = other.buf[i] & self.buf[i]
        result.buf[self.buf_size - 1] &= self.final_bitmask
        return result

    def logical_and(self, bitarray other, bitarray result = None):
        return self._logical_and(other, result)

    def __and__(self, bitarray other):
        # Wrap it directly here.
        return self.logical_and(other)

    def __iand__(self, bitarray other):
        rv = self.logical_and(other, self)
        return rv

    cdef bitarray _logical_or(self, bitarray other, bitarray result = None):
        if other.size != self.size:
            raise IndexError
        if result is None:
            result = bitarray(self.size)
        for i in range(self.buf_size):
            result.buf[i] = other.buf[i] | self.buf[i]
        result.buf[self.buf_size - 1] &= self.final_bitmask
        return result

    def logical_or(self, bitarray other, bitarray result = None):
        return self._logical_or(other, result)

    def __or__(self, bitarray other):
        return self.logical_or(other)

    def __ior__(self, bitarray other):
        return self.logical_or(other, self)

    cdef bitarray _logical_xor(self, bitarray other, bitarray result = None):
        if other.size != self.size:
            raise IndexError
        if result is None:
            result = bitarray(self.size)
        for i in range(self.buf_size):
            result.buf[i] = other.buf[i] ^ self.buf[i]
        result.buf[self.buf_size - 1] &= self.final_bitmask
        return result

    def logical_xor(self, bitarray other, bitarray result = None):
        return self._logical_xor(other, result)

    def __xor__(self, bitarray other):
        return self.logical_xor(other)

    def __ixor__(self, bitarray other):
        return self.logical_xor(other, self)



# distutils: libraries = STD_LIBS
# distutils: include_dirs = LIB_DIR
# distutils: language = c++
# distutils: extra_compile_args = CPP14_FLAG OMP_ARGS
# distutils: extra_link_args = CPP14_FLAG OMP_ARGS
cimport cython

import numpy as np

cimport numpy as np
from libc.math cimport fabs
from libc.stdlib cimport free, malloc

from cython.parallel import parallel, prange

from yt.utilities.lib.element_mappings cimport (
    ElementSampler,
    P1Sampler3D,
    Q1Sampler3D,
    S2Sampler3D,
    Tet2Sampler3D,
    W1Sampler3D,
)
from yt.utilities.lib.primitives cimport (
    BBox,
    Patch,
    Ray,
    TetPatch,
    Triangle,
    patch_bbox,
    patch_centroid,
    ray_bbox_intersect,
    ray_patch_intersect,
    ray_tet_patch_intersect,
    ray_triangle_intersect,
    tet_patch_bbox,
    tet_patch_centroid,
    triangle_bbox,
    triangle_centroid,
)

from .image_samplers cimport ImageSampler


cdef ElementSampler Q1Sampler = Q1Sampler3D()
cdef ElementSampler P1Sampler = P1Sampler3D()
cdef ElementSampler W1Sampler = W1Sampler3D()
cdef ElementSampler S2Sampler = S2Sampler3D()
cdef ElementSampler Tet2Sampler = Tet2Sampler3D()

cdef extern from "platform_dep.h" nogil:
    double fmax(double x, double y)
    double fmin(double x, double y)

# define some constants
cdef np.float64_t INF = np.inf
cdef np.int64_t   LEAF_SIZE = 16


cdef class BVH:
    '''

    This class implements a bounding volume hierarchy (BVH), a spatial acceleration
    structure for fast ray-tracing. A BVH is like a kd-tree, except that instead of
    partitioning the *volume* of the parent to create the children, we partition the
    primitives themselves into 'left' or 'right' sub-trees. The bounding volume for a
    node is then determined by computing the bounding volume of the primitives that
    belong to it. This allows us to quickly discard primitives that are not close
    to intersecting a given ray.

    This class is currently used to provide software 3D rendering support for
    finite element datasets. For 1st-order meshes, every element of the mesh is
    triangulated, and this set of triangles forms the primitives that will be used
    for the ray-trace. The BVH can then quickly determine which element is hit by
    each ray associated with the image plane, and the appropriate interpolation can
    be performed to sample the finite element solution at that hit position.

    Currently, 2nd-order meshes are only supported for 20-node hexahedral elements.
    There, the primitive type is a bi-quadratic patch instead of a triangle, and
    each intersection involves computing a Newton-Raphson solve.

    See yt/utilities/lib/primitives.pyx for the definitions of both of these primitive
    types.

    '''

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def __cinit__(self,
                  np.float64_t[:, :] vertices,
                  np.int64_t[:, :] indices,
                  np.float64_t[:, :] field_data):

        self.num_elem = indices.shape[0]
        self.num_verts_per_elem = indices.shape[1]
        self.num_field_per_elem = field_data.shape[1]

        # We need to figure out what kind of elements we've been handed.
        if self.num_verts_per_elem == 8:
            self.num_prim_per_elem = HEX_NT
            self.tri_array = triangulate_hex
            self.sampler = Q1Sampler
        elif self.num_verts_per_elem == 6:
            self.num_prim_per_elem = WEDGE_NT
            self.tri_array = triangulate_wedge
            self.sampler = W1Sampler
        elif self.num_verts_per_elem == 4:
            self.num_prim_per_elem = TETRA_NT
            self.tri_array = triangulate_tetra
            self.sampler = P1Sampler
        elif self.num_verts_per_elem == 20:
            self.num_prim_per_elem = 6
            self.sampler = S2Sampler
        elif self.num_verts_per_elem == 10:
            self.num_prim_per_elem = 4
            self.sampler = Tet2Sampler
        else:
            raise NotImplementedError("Could not determine element type for "
                                      "nverts = %d. " % self.num_verts_per_elem)
        self.num_prim = self.num_prim_per_elem*self.num_elem

        # allocate storage
        cdef np.int64_t v_size = self.num_verts_per_elem * self.num_elem * 3
        self.vertices = <np.float64_t*> malloc(v_size * sizeof(np.float64_t))
        cdef np.int64_t f_size = self.num_field_per_elem * self.num_elem
        self.field_data = <np.float64_t*> malloc(f_size * sizeof(np.float64_t))
        self.prim_ids = <np.int64_t*> malloc(self.num_prim * sizeof(np.int64_t))
        self.centroids = <np.float64_t**> malloc(self.num_prim * sizeof(np.float64_t*))
        cdef np.int64_t i
        for i in range(self.num_prim):
            self.centroids[i] = <np.float64_t*> malloc(3*sizeof(np.float64_t))
        self.bboxes = <BBox*> malloc(self.num_prim * sizeof(BBox))

        # create data buffers
        cdef np.int64_t j, k
        cdef np.int64_t field_offset, vertex_offset
        for i in range(self.num_elem):
            for j in range(self.num_verts_per_elem):
                vertex_offset = i*self.num_verts_per_elem*3 + j*3
                for k in range(3):
                    self.vertices[vertex_offset + k] = vertices[indices[i,j]][k]
            field_offset = i*self.num_field_per_elem
            for j in range(self.num_field_per_elem):
                self.field_data[field_offset + j] = field_data[i][j]

        # set up primitives
        if self.num_verts_per_elem == 20:
            self.primitives = malloc(self.num_prim * sizeof(Patch))
            self.get_centroid = patch_centroid
            self.get_bbox = patch_bbox
            self.get_intersect = ray_patch_intersect
            self._set_up_patches(vertices, indices)
        elif self.num_verts_per_elem == 10:
            self.primitives = malloc(self.num_prim * sizeof(TetPatch))
            self.get_centroid = tet_patch_centroid
            self.get_bbox = tet_patch_bbox
            self.get_intersect = ray_tet_patch_intersect
            self._set_up_tet_patches(vertices, indices)
        else:
            self.primitives = malloc(self.num_prim * sizeof(Triangle))
            self.get_centroid = triangle_centroid
            self.get_bbox = triangle_bbox
            self.get_intersect = ray_triangle_intersect
            self._set_up_triangles(vertices, indices)

        self.root = self._recursive_build(0, self.num_prim)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _set_up_patches(self, np.float64_t[:, :] vertices,
                              np.int64_t[:, :] indices) noexcept nogil:
        cdef Patch* patch
        cdef np.int64_t i, j, k, ind, idim
        cdef np.int64_t offset, prim_index
        for i in range(self.num_elem):
            offset = self.num_prim_per_elem*i
            for j in range(self.num_prim_per_elem):  # for each face
                prim_index = offset + j
                patch = &( <Patch*> self.primitives)[prim_index]
                self.prim_ids[prim_index] = prim_index
                patch.elem_id = i
                for k in range(8):  # for each vertex
                    ind = hex20_faces[j][k]
                    for idim in range(3):  # for each spatial dimension (yikes)
                        patch.v[k][idim] = vertices[indices[i, ind]][idim]
                self.get_centroid(self.primitives,
                                  prim_index,
                                  self.centroids[prim_index])
                self.get_bbox(self.primitives,
                              prim_index,
                              &(self.bboxes[prim_index]))

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _set_up_tet_patches(self, np.float64_t[:, :] vertices,
                              np.int64_t[:, :] indices) noexcept nogil:
        cdef TetPatch* tet_patch
        cdef np.int64_t i, j, k, ind, idim
        cdef np.int64_t offset, prim_index
        for i in range(self.num_elem):
            offset = self.num_prim_per_elem*i
            for j in range(self.num_prim_per_elem):  # for each face
                prim_index = offset + j
                tet_patch = &( <TetPatch*> self.primitives)[prim_index]
                self.prim_ids[prim_index] = prim_index
                tet_patch.elem_id = i
                for k in range(6):  # for each vertex
                    ind = tet10_faces[j][k]
                    for idim in range(3):  # for each spatial dimension (yikes)
                        tet_patch.v[k][idim] = vertices[indices[i, ind]][idim]
                self.get_centroid(self.primitives,
                                  prim_index,
                                  self.centroids[prim_index])
                self.get_bbox(self.primitives,
                              prim_index,
                              &(self.bboxes[prim_index]))

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _set_up_triangles(self, np.float64_t[:, :] vertices,
                                np.int64_t[:, :] indices) noexcept nogil:
        # fill our array of primitives
        cdef np.int64_t offset, tri_index
        cdef np.int64_t v0, v1, v2
        cdef Triangle* tri
        cdef np.int64_t i, j, k
        for i in range(self.num_elem):
            offset = self.num_prim_per_elem*i
            for j in range(self.num_prim_per_elem):
                tri_index = offset + j
                self.prim_ids[tri_index] = tri_index
                tri = &(<Triangle*> self.primitives)[tri_index]
                tri.elem_id = i
                v0 = indices[i][self.tri_array[j][0]]
                v1 = indices[i][self.tri_array[j][1]]
                v2 = indices[i][self.tri_array[j][2]]
                for k in range(3):
                    tri.p0[k] = vertices[v0][k]
                    tri.p1[k] = vertices[v1][k]
                    tri.p2[k] = vertices[v2][k]
                self.get_centroid(self.primitives,
                                  tri_index,
                                  self.centroids[tri_index])
                self.get_bbox(self.primitives,
                              tri_index,
                              &(self.bboxes[tri_index]))

    cdef void _recursive_free(self, BVHNode* node) noexcept nogil:
        if node.end - node.begin > LEAF_SIZE:
            self._recursive_free(node.left)
            self._recursive_free(node.right)
        free(node)

    def __dealloc__(self):
        if self.root == NULL:
            return
        self._recursive_free(self.root)
        free(self.primitives)
        free(self.prim_ids)
        for i in range(self.num_prim):
            free(self.centroids[i])
        free(self.centroids)
        free(self.bboxes)
        free(self.field_data)
        free(self.vertices)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef np.int64_t _partition(self, np.int64_t begin, np.int64_t end,
                               np.int64_t ax, np.float64_t split) noexcept nogil:
        # this re-orders the primitive array so that all of the primitives
        # to the left of mid have centroids less than or equal to "split"
        # along the direction "ax". All the primitives to the right of mid
        # will have centroids *greater* than "split" along "ax".
        cdef np.int64_t mid = begin
        while (begin != end):
            if self.centroids[mid][ax] > split:
                mid += 1
            elif self.centroids[begin][ax] > split:
                self.prim_ids[mid], self.prim_ids[begin] = \
                self.prim_ids[begin], self.prim_ids[mid]
                self.centroids[mid], self.centroids[begin] = \
                self.centroids[begin], self.centroids[mid]
                self.bboxes[mid], self.bboxes[begin] = \
                self.bboxes[begin], self.bboxes[mid]
                mid += 1
            begin += 1
        return mid

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _get_node_bbox(self, BVHNode* node,
                             np.int64_t begin, np.int64_t end) noexcept nogil:
        cdef np.int64_t i, j
        cdef BBox box = self.bboxes[begin]
        for i in range(begin+1, end):
            for j in range(3):
                box.left_edge[j] = fmin(box.left_edge[j],
                                        self.bboxes[i].left_edge[j])
                box.right_edge[j] = fmax(box.right_edge[j],
                                         self.bboxes[i].right_edge[j])
        node.bbox = box

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void intersect(self, Ray* ray) noexcept nogil:
        self._recursive_intersect(ray, self.root)

        if ray.elem_id < 0:
            return

        cdef np.float64_t[3] position
        cdef np.int64_t i
        for i in range(3):
            position[i] = ray.origin[i] + ray.t_far*ray.direction[i]

        cdef np.float64_t* vertex_ptr
        cdef np.float64_t* field_ptr
        vertex_ptr = self.vertices + ray.elem_id*self.num_verts_per_elem*3
        field_ptr = self.field_data + ray.elem_id*self.num_field_per_elem

        cdef np.float64_t[4] mapped_coord
        self.sampler.map_real_to_unit(mapped_coord, vertex_ptr, position)
        if self.num_field_per_elem == 1:
            ray.data_val = field_ptr[0]
        else:
            ray.data_val = self.sampler.sample_at_unit_point(mapped_coord,
                                                             field_ptr)
        ray.near_boundary = self.sampler.check_mesh_lines(mapped_coord)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _recursive_intersect(self, Ray* ray, BVHNode* node) noexcept nogil:

        # check for bbox intersection:
        if not ray_bbox_intersect(ray, node.bbox):
            return

        # check for leaf
        cdef np.int64_t i
        if (node.end - node.begin) <= LEAF_SIZE:
            for i in range(node.begin, node.end):
                self.get_intersect(self.primitives, self.prim_ids[i], ray)
            return

        # if not leaf, intersect with left and right children
        self._recursive_intersect(ray, node.left)
        self._recursive_intersect(ray, node.right)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef BVHNode* _recursive_build(self, np.int64_t begin, np.int64_t end) noexcept nogil:
        cdef BVHNode *node = <BVHNode* > malloc(sizeof(BVHNode))
        node.begin = begin
        node.end = end

        self._get_node_bbox(node, begin, end)

        # check for leaf
        if (end - begin) <= LEAF_SIZE:
            return node

        # we use the "split in the middle of the longest axis approach"
        # see: http://www.vadimkravcenko.com/bvh-tree-building/

        # compute longest dimension
        cdef np.int64_t ax = 0
        cdef np.float64_t d = fabs(node.bbox.right_edge[0] -
                                   node.bbox.left_edge[0])
        if fabs(node.bbox.right_edge[1] - node.bbox.left_edge[1]) > d:
            ax = 1
        if fabs(node.bbox.right_edge[2] - node.bbox.left_edge[2]) > d:
            ax = 2

        # split in half along that dimension
        cdef np.float64_t split = 0.5*(node.bbox.right_edge[ax] +
                                       node.bbox.left_edge[ax])

        # sort triangle list
        cdef np.int64_t mid = self._partition(begin, end, ax, split)

        if(mid == begin or mid == end):
            mid = begin + (end-begin)/2

        # recursively build sub-trees
        node.left = self._recursive_build(begin, mid)
        node.right = self._recursive_build(mid, end)

        return node


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void cast_rays(np.float64_t* image,
                    const np.float64_t* origins,
                    const np.float64_t* direction,
                    const int N,
                    BVH bvh) noexcept nogil:

    cdef Ray* ray
    cdef int i, j, k

    with nogil, parallel():

        ray = <Ray *> malloc(sizeof(Ray))

        for k in range(3):
            ray.direction[k] = direction[k]
            ray.inv_dir[k] = 1.0 / direction[k]

        for i in prange(N):
            for j in range(3):
                ray.origin[j] = origins[N*j + i]
            ray.t_far = INF
            ray.t_near = 0.0
            ray.data_val = 0
            ray.elem_id = -1
            bvh.intersect(ray)
            image[i] = ray.data_val

        free(ray)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_ray_trace(np.ndarray[np.float64_t, ndim=1] image,
                   np.ndarray[np.float64_t, ndim=2] origins,
                   np.ndarray[np.float64_t, ndim=1] direction,
                   BVH bvh):

    cdef int N = origins.shape[0]
    cast_rays(&image[0], &origins[0, 0], &direction[0], N, bvh)

cdef class BVHMeshSampler(ImageSampler):

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def __call__(self,
                 BVH bvh,
                 int num_threads = 0):
        '''

        This function is supposed to cast the rays and return the
        image.

        '''

        cdef int vi, vj, i, j
        cdef np.float64_t *v_pos
        cdef np.float64_t *v_dir
        cdef np.int64_t nx, ny, size
        cdef np.float64_t width[3]
        for i in range(3):
            width[i] = self.width[i]
        nx = self.nv[0]
        ny = self.nv[1]
        size = nx * ny
        cdef Ray* ray
        with nogil, parallel():
            ray = <Ray *> malloc(sizeof(Ray))
            v_pos = <np.float64_t *> malloc(3 * sizeof(np.float64_t))
            v_dir = <np.float64_t *> malloc(3 * sizeof(np.float64_t))
            for j in prange(size):
                vj = j % ny
                vi = (j - vj) / ny
                vj = vj
                self.vector_function(self, vi, vj, width, v_dir, v_pos)
                for i in range(3):
                    ray.origin[i] = v_pos[i]
                    ray.direction[i] = v_dir[i]
                    ray.inv_dir[i] = 1.0 / v_dir[i]
                ray.t_far = 1e37
                ray.t_near = 0.0
                ray.data_val = 0
                ray.elem_id = -1
                bvh.intersect(ray)
                self.image[vi, vj, 0] = ray.data_val
                self.image_used[vi, vj] = ray.elem_id
                self.mesh_lines[vi, vj] = ray.near_boundary
                self.zbuffer[vi, vj] = ray.t_far
            free(v_pos)
            free(v_dir)
            free(ray)



from numpy.testing import assert_equal

from yt.data_objects.profiles import create_profile
from yt.testing import fake_random_ds
from yt.visualization.plot_window import ProjectionPlot, SlicePlot
from yt.visualization.profile_plotter import PhasePlot, ProfilePlot


def test_field_access():
    ds = fake_random_ds(16)

    ad = ds.all_data()
    sp = ds.sphere(ds.domain_center, 0.25)
    cg = ds.covering_grid(0, ds.domain_left_edge, ds.domain_dimensions)
    scg = ds.smoothed_covering_grid(0, ds.domain_left_edge, ds.domain_dimensions)
    sl = ds.slice(0, ds.domain_center[0])
    proj = ds.proj(("gas", "density"), 0)
    prof = create_profile(ad, ("index", "radius"), ("gas", "density"))

    for data_object in [ad, sp, cg, scg, sl, proj, prof]:
        assert_equal(data_object["gas", "density"], data_object[ds.fields.gas.density])

    for field in [("gas", "density"), ds.fields.gas.density]:
        ad = ds.all_data()
        prof = ProfilePlot(ad, ("index", "radius"), field)
        phase = PhasePlot(ad, ("index", "radius"), field, ("gas", "cell_mass"))
        s = SlicePlot(ds, 2, field)
        oas = SlicePlot(ds, [1, 1, 1], field)
        p = ProjectionPlot(ds, 2, field)
        oap = ProjectionPlot(ds, [1, 1, 1], field)

        for plot_object in [s, oas, p, oap, prof, phase]:
            plot_object.render()
            if hasattr(plot_object, "_frb"):
                plot_object._frb[field]



import os
import re

import pytest
from PIL import Image

import yt
from yt.testing import fake_amr_ds

# avoid testing every supported format as some backends may be buggy on testing platforms
FORMATS_TO_TEST = [".eps", ".pdf", ".svg"]


@pytest.fixture(scope="session")
def simple_sliceplot():
    ds = fake_amr_ds()
    p = yt.SlicePlot(ds, "z", "Density")
    yield p


def test_save_to_path(simple_sliceplot, tmp_path):
    p = simple_sliceplot
    p.save(f"{tmp_path}/")
    assert len(list((tmp_path).glob("*.png"))) == 1


def test_metadata(simple_sliceplot, tmp_path):
    simple_sliceplot.save(tmp_path / "ala.png")
    with Image.open(tmp_path / "ala.png") as img:
        assert "Software" in img.info
        assert "yt-" in img.info["Software"]
    simple_sliceplot.save(tmp_path / "ala.pdf")
    with open(tmp_path / "ala.pdf", "rb") as f:
        assert b"|yt-" in f.read()


def test_save_to_missing_path(simple_sliceplot, tmp_path):
    # the missing layer should be created
    p = simple_sliceplot

    # using forward slashes should work even on windows !
    save_path = os.path.join(tmp_path / "out") + "/"
    p.save(save_path)
    assert os.path.exists(save_path)
    assert len(list((tmp_path / "out").glob("*.png"))) == 1


def test_save_to_missing_path_with_file_prefix(simple_sliceplot, tmp_path):
    # see issue
    # https://github.com/yt-project/yt/issues/3210
    p = simple_sliceplot
    p.save(tmp_path.joinpath("out", "saymyname"))
    assert (tmp_path / "out").exists()
    output_files = list((tmp_path / "out").glob("*.png"))
    assert len(output_files) == 1
    assert output_files[0].stem.startswith("saymyname")  # you're goddamn right


@pytest.mark.parametrize("ext", FORMATS_TO_TEST)
def test_suffix_from_filename(ext, simple_sliceplot, tmp_path):
    p = simple_sliceplot

    target = (tmp_path / "myfile").with_suffix(ext)
    # this shouldn't raise a warning, see issue
    # https://github.com/yt-project/yt/issues/3667
    p.save(target)
    assert target.is_file()


@pytest.mark.parametrize("ext", FORMATS_TO_TEST)
def test_suffix_clashing(ext, simple_sliceplot, tmp_path):
    if ext == ".png":
        pytest.skip()

    p = simple_sliceplot

    target = (tmp_path / "myfile").with_suffix(ext)
    expected_warning = re.compile(
        rf"Received two valid image formats {ext.removeprefix('.')!r} "
        r"\(from filename\) and 'png' \(from suffix\)\. The former is ignored\."
    )

    with pytest.warns(UserWarning, match=expected_warning):
        p.save(target, suffix="png")
    output_files = list(tmp_path.glob("*.png"))
    assert len(output_files) == 1
    assert output_files[0].stem.startswith("myfile")
    assert not list((tmp_path / "out").glob(f"*.{ext}"))


def test_invalid_format_from_filename(simple_sliceplot, tmp_path):
    p = simple_sliceplot
    target = (tmp_path / "myfile").with_suffix(".nope")

    p.save(target)
    output_files = list(tmp_path.glob("*"))
    assert len(output_files) == 1
    # the output filename may contain a generated part
    # it's not exactly clear if it's desirable or intended in this case
    # so we just check conditions that should hold in any case
    assert output_files[0].name.startswith("myfile.nope")
    assert output_files[0].name.endswith(".png")


def test_invalid_format_from_suffix(simple_sliceplot, tmp_path):
    p = simple_sliceplot
    target = tmp_path / "myfile"
    with pytest.raises(ValueError, match=r"Unsupported file format 'nope'"):
        p.save(target, suffix="nope")



cimport cython
cimport pyembree.rtcore as rtc
cimport pyembree.rtcore_geometry as rtcg
cimport pyembree.rtcore_ray as rtcr

from .mesh_construction cimport Patch, Tet_Patch


cdef void patchIntersectFunc(Patch* patches,
                             rtcr.RTCRay& ray,
                             size_t item) noexcept nogil

cdef void patchBoundsFunc(Patch* patches,
                          size_t item,
                          rtcg.RTCBounds* bounds_o) noexcept nogil

cdef void tet_patchIntersectFunc(Tet_Patch* tet_patches,
                             rtcr.RTCRay& ray,
                             size_t item) noexcept nogil

cdef void tet_patchBoundsFunc(Tet_Patch* tet_patches,
                          size_t item,
                          rtcg.RTCBounds* bounds_o) noexcept nogil



"""
CF-Radial-specific IO functions



"""

import numpy as np

from yt.utilities.io_handler import BaseIOHandler


class CFRadialIOHandler(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "cf_radial"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        # This needs to allocate a set of arrays inside a dictionary, where the
        # keys are the (ftype, fname) tuples and the values are arrays that
        # have been masked using whatever selector method is appropriate.  The
        # dict gets returned at the end and it should be flat, with selected
        # data.

        rv = {field: np.empty(size, dtype="float64") for field in fields}

        offset = 0

        with self.ds._handle() as xr_ds_handle:
            for field in fields:
                for chunk in chunks:
                    for grid in chunk.objs:
                        variable = xr_ds_handle.variables[field[1]]
                        data = variable.values[0, ...].T
                        offset += grid.select(selector, data, rv[field], offset)
        return rv



class ConfigNode:
    def __init__(self, key, parent=None):
        self.key = key
        self.children = {}
        self.parent = parent

    def add(self, key, child):
        self.children[key] = child
        child.parent = self

    def update(self, other, extra_data=None):
        def _recursive_upsert(other_dict, keys):
            for key, val in other_dict.items():
                new_keys = keys + [key]
                if isinstance(val, dict):
                    _recursive_upsert(val, new_keys)
                else:
                    self.upsert_from_list(new_keys, val, extra_data)

        _recursive_upsert(other, keys=[])

    def get_child(self, key, constructor=None):
        if key in self.children:
            child = self.children[key]
        elif constructor is not None:
            child = self.children[key] = constructor()
        else:
            raise KeyError(f"Cannot get key {key}")
        return child

    def add_child(self, key):
        self.get_child(key, lambda: ConfigNode(key, parent=self))

    def remove_child(self, key):
        self.children.pop(key)

    def upsert_from_list(self, keys, value, extra_data=None):
        key, *next_keys = keys
        if len(next_keys) == 0:  # reach the end of the upsert
            leaf = self.get_child(
                key,
                lambda: ConfigLeaf(
                    key, parent=self, value=value, extra_data=extra_data
                ),
            )
            leaf.value = value
            leaf.extra_data = extra_data
            if not isinstance(leaf, ConfigLeaf):
                raise RuntimeError(f"Expected a ConfigLeaf, got {leaf}!")
        else:
            next_node = self.get_child(key, lambda: ConfigNode(key, parent=self))
            if not isinstance(next_node, ConfigNode):
                raise RuntimeError(f"Expected a ConfigNode, got {next_node}!")
            next_node.upsert_from_list(next_keys, value, extra_data)

    def get_from_list(self, key_list):
        next, *key_list_remainder = key_list
        child = self.get_child(next)
        if len(key_list_remainder) == 0:
            return child
        else:
            return child.get_from_list(key_list_remainder)

    def get(self, *keys):
        return self.get_from_list(keys)

    def get_leaf(self, *keys, callback=lambda leaf: leaf.value):
        leaf = self.get_from_list(keys)
        return callback(leaf)

    def pop_leaf(self, keys):
        *node_keys, leaf_key = keys
        node = self.get_from_list(node_keys)
        node.children.pop(leaf_key)

    def get_deepest_leaf(self, *keys, callback=lambda leaf: leaf.value):
        root_key, *keys, leaf_key = keys

        root_node = self.get_child(root_key)
        node_list = [root_node]
        node = root_node

        # Traverse the tree down following the keys
        for k in keys:
            try:
                node = node.get_child(k)
                node_list.append(node)
            except KeyError:
                break

        # For each node, starting from the deepest, try to find the leaf
        for node in reversed(node_list):
            try:
                leaf = node.get_child(leaf_key)
                if not isinstance(leaf, ConfigLeaf):
                    raise RuntimeError(f"Expected a ConfigLeaf, got {leaf}!")
                return callback(leaf)
            except KeyError:
                continue
        raise KeyError(f"Cannot any node that contains the leaf {leaf_key}.")

    def serialize(self):
        retval = {}
        for key, child in self.children.items():
            retval[key] = child.serialize()
        return retval

    @staticmethod
    def from_dict(other, parent=None, **kwa):
        me = ConfigNode(None, parent=parent)
        for key, val in other.items():
            if isinstance(val, dict):
                me.add(key, ConfigNode.from_dict(val, parent=me, **kwa))
            else:
                me.add(key, ConfigLeaf(key, parent=me, value=val, **kwa))
        return me

    def _as_dict_with_count(self, callback):
        data = {}
        total_count = 0
        for key, child in self.children.items():
            if isinstance(child, ConfigLeaf):
                total_count += 1
                data[key] = callback(child)
            elif isinstance(child, ConfigNode):
                child_data, count = child._as_dict_with_count(callback)
                total_count += count
                if count > 0:
                    data[key] = child_data

        return data, total_count

    def as_dict(self, callback=lambda child: child.value):
        data, _ = self._as_dict_with_count(callback)
        return data

    def __repr__(self):
        return f"<Node {self.key}>"

    def __contains__(self, item):
        return item in self.children

    # Add support for IPython rich display
    # see https://ipython.readthedocs.io/en/stable/config/integrating.html
    def _repr_json_(self):
        return self.as_dict()


class ConfigLeaf:
    def __init__(self, key, parent: ConfigNode, value, extra_data=None):
        self.key = key  # the name of the config leaf
        self._value = value
        self.parent = parent
        self.extra_data = extra_data

    def serialize(self):
        return self.value

    def get_tree(self):
        node = self
        parents = []
        while node is not None:
            parents.append(node)
            node = node.parent

        return reversed(parents)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        if type(self.value) is type(new_value):
            self._value = new_value
        else:
            tree = self.get_tree()
            tree_str = ".".join(node.key for node in tree if node.key)
            msg = f"Error when setting {tree_str}.\n"
            msg += (
                "Tried to assign a value of type "
                f"{type(new_value)}, expected type {type(self.value)}."
            )
            source = self.extra_data.get("source", None)
            if source:
                msg += f"\nThis entry was last modified in {source}."
            raise TypeError(msg)

    def __repr__(self):
        return f"<Leaf {self.key}: {self.value}>"



import json
import time
from functools import cached_property

import numpy as np

from yt.data_objects.static_output import ParticleDataset, ParticleFile
from yt.frontends.sph.fields import SPHFieldInfo
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities.on_demand_imports import _requests as requests


class HTTPParticleFile(ParticleFile):
    pass


class HTTPStreamDataset(ParticleDataset):
    _load_requirements = ["requests"]
    _index_class = ParticleIndex
    _file_class = HTTPParticleFile
    _field_info_class = SPHFieldInfo
    _particle_mass_name = "Mass"
    _particle_coordinates_name = "Coordinates"
    _particle_velocity_name = "Velocities"
    filename_template = ""

    def __init__(
        self,
        base_url,
        dataset_type="http_particle_stream",
        unit_system="cgs",
        index_order=None,
        index_filename=None,
    ):
        self.base_url = base_url
        super().__init__(
            "",
            dataset_type=dataset_type,
            unit_system=unit_system,
            index_order=index_order,
            index_filename=index_filename,
        )

    def __str__(self):
        return self.base_url

    @cached_property
    def unique_identifier(self) -> str:
        return str(self.parameters.get("unique_identifier", time.time()))

    def _parse_parameter_file(self):
        self.dimensionality = 3
        self.refine_by = 2
        self.parameters["HydroMethod"] = "sph"

        # Here's where we're going to grab the JSON index file
        hreq = requests.get(self.base_url + "/yt_index.json")
        if hreq.status_code != 200:
            raise RuntimeError
        header = json.loads(hreq.content)
        header["particle_count"] = {
            int(k): header["particle_count"][k] for k in header["particle_count"]
        }
        self.parameters = header

        # Now we get what we need
        self.domain_left_edge = np.array(header["domain_left_edge"], "float64")
        self.domain_right_edge = np.array(header["domain_right_edge"], "float64")
        self.domain_dimensions = np.ones(3, "int32")
        self._periodicity = (True, True, True)

        self.current_time = header["current_time"]
        self.cosmological_simulation = int(header["cosmological_simulation"])
        for attr in (
            "current_redshift",
            "omega_lambda",
            "omega_matter",
            "hubble_constant",
        ):
            setattr(self, attr, float(header[attr]))

        self.file_count = header["num_files"]

    def _set_units(self):
        length_unit = float(self.parameters["units"]["length"])
        time_unit = float(self.parameters["units"]["time"])
        mass_unit = float(self.parameters["units"]["mass"])
        density_unit = mass_unit / length_unit**3
        velocity_unit = length_unit / time_unit
        self._unit_base = {}
        self._unit_base["cm"] = 1.0 / length_unit
        self._unit_base["s"] = 1.0 / time_unit
        super()._set_units()
        self.conversion_factors["velocity"] = velocity_unit
        self.conversion_factors["mass"] = mass_unit
        self.conversion_factors["density"] = density_unit

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if not filename.startswith("http://"):
            return False

        if cls._missing_load_requirements():
            return False

        return requests.get(filename + "/yt_index.json").status_code == 200



import numpy as np

from yt.utilities.physical_ratios import (
    boltzmann_constant_erg_per_K,
    cm_per_mpc,
    mass_hydrogen_grams,
    newton_cgs,
    rho_crit_g_cm3_h2,
)


def cosmology_get_units(
    hubble_constant, omega_matter, box_size, initial_redshift, current_redshift
):
    """
    Return a dict of Enzo cosmological unit conversions.
    """

    zp1 = 1.0 + current_redshift
    zip1 = 1.0 + initial_redshift

    k = {}
    # For better agreement with values calculated by Enzo,
    # adopt the exact constants that are used there.

    time_scaling = np.sqrt(1 / (4 * np.pi * newton_cgs * rho_crit_g_cm3_h2))
    vel_scaling = cm_per_mpc / time_scaling
    temp_scaling = mass_hydrogen_grams / boltzmann_constant_erg_per_K * vel_scaling**2

    k["utim"] = time_scaling / np.sqrt(omega_matter) / hubble_constant / zip1**1.5
    k["urho"] = rho_crit_g_cm3_h2 * omega_matter * hubble_constant**2 * zp1**3
    k["uxyz"] = cm_per_mpc * box_size / hubble_constant / zp1
    k["uaye"] = 1.0 / zip1
    k["uvel"] = vel_scaling * box_size * np.sqrt(omega_matter) * np.sqrt(zip1)
    k["utem"] = temp_scaling * (box_size**2) * omega_matter * zip1
    k["aye"] = zip1 / zp1
    return k



"""
Chimera-specific fields



"""

from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer

# We need to specify which fields we might have in our dataset.  The field info
# container subclass here will define which fields it knows about.  There are
# optionally methods on it that get called which can be subclassed.


class ChimeraFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("e_int", ("erg", ["Internal Energy"], "Internal Energy")),
        ("entropy", ("", ["Entropy"], None)),
        ("rho_c", ("g/cm**3", ["density", "Density"], "Density")),
        ("dudt_nu", ("erg/s", [], None)),
        ("dudt_nuc", ("erg/s", [], None)),
        ("grav_x_c", ("cm/s**2", [], None)),
        ("grav_y_c", ("cm/s**2", [], None)),
        ("grav_z_c", ("cm/s**2", [], None)),
        ("press", ("erg/cm**3", ["pressure"], "Pressure")),
        ("t_c", ("K", ["temperature"], "Temperature")),
        ("u_c", ("cm/s", ["v_radial"], "Radial Velocity")),
        ("v_c", ("cm/s", ["v_theta"], "Theta Velocity")),
        ("v_csound", ("", [], None)),
        ("wBVMD", ("1/s", [], "BruntViasala_freq")),
        ("w_c", ("cm/s", ["v_phi"], "Phi Velocity")),
        ("ye_c", ("", [], None)),
        ("ylep", ("", [], None)),
        ("a_nuc_rep_c", ("", [], None)),
        ("be_nuc_rep_c", ("", [], None)),
        ("e_book", ("", [], None)),
        ("nse_c", ("", [], None)),
        ("z_nuc_rep_c", ("", [], None)),
    )
    # Each entry here is of the form
    # ( "name", ("units", ["fields", "to", "alias"], # "display_name")),

    known_particle_fields = (
        # Identical form to above
    )

    def __init__(self, ds, field_list):
        super().__init__(ds, field_list)
        # If you want, you can check self.field_list

    def setup_fluid_fields(self):
        # Here we do anything that might need info about the dataset.
        # You can use self.alias, self.add_output_field (for on-disk fields)
        # and self.add_field (for derived fields).

        def _test(field, data):
            return data["chimera", "rho_c"]

        self.add_field(
            ("chimera", "test"), sampling_type="cell", function=_test, units="g/cm**3"
        )

    def setup_particle_fields(self, ptype):
        super().setup_particle_fields(ptype)
        # This will get called for every particle type.



import matplotlib.pyplot as plt

import yt

ds = yt.load("GasSloshing/sloshing_nomag2_hdf5_plt_cnt_0150")

# Get the first sphere
sp0 = ds.sphere(ds.domain_center, (500.0, "kpc"))

# Compute the bulk velocity from the cells in this sphere
bulk_vel = sp0.quantities.bulk_velocity()


# Get the second sphere
sp1 = ds.sphere(ds.domain_center, (500.0, "kpc"))

# Set the bulk velocity field parameter
sp1.set_field_parameter("bulk_velocity", bulk_vel)

# Radial profile without correction

rp0 = yt.create_profile(
    sp0,
    ("index", "radius"),
    ("gas", "radial_velocity"),
    units={("index", "radius"): "kpc"},
    logs={("index", "radius"): False},
)

# Radial profile with correction for bulk velocity

rp1 = yt.create_profile(
    sp1,
    ("index", "radius"),
    ("gas", "radial_velocity"),
    units={("index", "radius"): "kpc"},
    logs={("index", "radius"): False},
)

# Make a plot using matplotlib

fig = plt.figure()
ax = fig.add_subplot(111)

ax.plot(
    rp0.x.value,
    rp0["gas", "radial_velocity"].in_units("km/s").value,
    rp1.x.value,
    rp1["gas", "radial_velocity"].in_units("km/s").value,
)

ax.set_xlabel(r"$\mathrm{r\ (kpc)}$")
ax.set_ylabel(r"$\mathrm{v_r\ (km/s)}$")
ax.legend(["Without Correction", "With Correction"])

fig.savefig(f"{ds}_profiles.png")



import numpy as np

from yt.loaders import load_uniform_grid
from yt.visualization.plot_window import ProjectionPlot


def test_ds_arr_invariance_under_projection_plot(tmp_path):
    data_array = np.random.random((10, 10, 10))
    bbox = np.array([[-100, 100], [-100, 100], [-100, 100]])
    data = {("gas", "density"): (data_array, "g*cm**(-3)")}
    ds = load_uniform_grid(data, data_array.shape, length_unit="kpc", bbox=bbox)

    start_source = np.array((0, 0, -0.5))
    end_source = np.array((0, 0, 0.5))
    start = ds.arr(start_source, "unitary")
    end = ds.arr(end_source, "unitary")

    start_i = start.copy()
    end_i = end.copy()

    p = ProjectionPlot(ds, 0, "number_density")
    p.annotate_line(start, end)
    p.save(tmp_path)

    # for lack of a unyt.testing.assert_unit_array_equal function
    np.testing.assert_array_equal(start_i, start)
    assert start_i.units == start.units
    np.testing.assert_array_equal(end_i, end)
    assert end_i.units == end.units



"""
Definitions for the traversal code




"""


import numpy as np

cimport cython
cimport numpy as np

from .image_samplers cimport ImageSampler
from .volume_container cimport VolumeContainer, vc_index, vc_pos_index

ctypedef void sampler_function(
                VolumeContainer *vc,
                np.float64_t v_pos[3],
                np.float64_t v_dir[3],
                np.float64_t enter_t,
                np.float64_t exit_t,
                int index[3],
                void *data) noexcept nogil

#-----------------------------------------------------------------------------
# walk_volume(VolumeContainer *vc,  np.float64_t v_pos[3], np.float64_t v_dir[3], sampler_function *sample,
#             void *data, np.float64_t *return_t = NULL, np.float64_t max_t = 1.0)
#    vc        VolumeContainer*  : Pointer to the volume container to be traversed.
#    v_pos     np.float64_t[3]   : The x,y,z coordinates of the ray's origin.
#    v_dir     np.float64_t[3]   : The x,y,z coordinates of the ray's direction.
#    sample    sampler_function* : Pointer to the sampler function to be used.
#    return_t  np.float64_t*     : Pointer to the final value of t that is still inside the volume container. Defaulted to NULL.
#    max_t     np.float64_t      : The maximum value of t that the ray is allowed to travel. Defaulted to 1.0 (no restriction).
#
#    Note: 't' is not time here. Rather, it is a factor representing the difference between the initial point 'v_pos'
#             and the end point, which we might call v_end. It is scaled such that v_pos + v * t = v_pos at t = 0.0, and
#             v_end at t = 1.0. Therefore, if max_t is set to 1.0, there is no restriction on t.
#
# Written by the yt Development Team.
# Encapsulates the Amanatides & Woo "Fast Traversal Voxel Algorithm" to walk over a volume container 'vc'
# The function occurs in two phases, initialization and traversal.
# See: https://www.researchgate.net/publication/2611491_A_Fast_Voxel_Traversal_Algorithm_for_Ray_Tracing
# Returns: The number of voxels hit during the traversal phase. If the traversal phase is not reached, returns 0.
#-----------------------------------------------------------------------------
cdef int walk_volume(VolumeContainer *vc,
                     np.float64_t v_pos[3],
                     np.float64_t v_dir[3],
                     sampler_function *sampler,
                     void *data,
                     np.float64_t *return_t = *,
                     np.float64_t max_t = *) noexcept nogil



import os
from collections import defaultdict
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from functools import _make_key, lru_cache

import numpy as np

from yt._typing import FieldKey, ParticleCoordinateTuple
from yt.geometry.selection_routines import GridSelector
from yt.utilities.on_demand_imports import _h5py as h5py

io_registry = {}

use_caching = 0


def _make_io_key(args, *_args, **kwargs):
    self, obj, field, ctx = args
    # Ignore self because we have a self-specific cache
    return _make_key((obj.id, field), *_args, **kwargs)


class BaseIOHandler:
    _vector_fields: dict[str, int] = {}
    _dataset_type: str
    _particle_reader = False
    _cache_on = False
    _misses = 0
    _hits = 0

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        if hasattr(cls, "_dataset_type"):
            io_registry[cls._dataset_type] = cls
        if use_caching and hasattr(cls, "_read_obj_field"):
            cls._read_obj_field = lru_cache(
                maxsize=use_caching, typed=True, make_key=_make_io_key
            )(cls._read_obj_field)

    def __init__(self, ds):
        self.queue = defaultdict(dict)
        self.ds = ds
        self._last_selector_id = None
        self._last_selector_counts = None
        self._array_fields = {}
        self._cached_fields = {}

    # We need a function for reading a list of sets
    # and a function for *popping* from a queue all the appropriate sets
    @contextmanager
    def preload(self, chunk, fields: list[FieldKey], max_size):
        yield self

    def peek(self, grid, field):
        return self.queue[grid.id].get(field, None)

    def push(self, grid, field, data):
        if grid.id in self.queue and field in self.queue[grid.id]:
            raise ValueError
        self.queue[grid][field] = data

    def _field_in_backup(self, grid, backup_file, field_name):
        if os.path.exists(backup_file):
            fhandle = h5py.File(backup_file, mode="r")
            g = fhandle["data"]
            grid_group = g["grid_%010i" % (grid.id - grid._id_offset)]
            if field_name in grid_group:
                return_val = True
            else:
                return_val = False
            fhandle.close()
            return return_val
        else:
            return False

    def _read_data_set(self, grid, field):
        # check backup file first. if field not found,
        # call frontend-specific io method
        backup_filename = grid.ds.backup_filename
        if not os.path.exists(backup_filename):
            return self._read_data(grid, field)
        elif self._field_in_backup(grid, backup_filename, field):
            fhandle = h5py.File(backup_filename, mode="r")
            g = fhandle["data"]
            grid_group = g["grid_%010i" % (grid.id - grid._id_offset)]
            data = grid_group[field][:]
            fhandle.close()
            return data
        else:
            return self._read_data(grid, field)

    # Now we define our interface
    def _read_data(self, grid, field):
        pass

    def _read_fluid_selection(
        self, chunks, selector, fields: list[FieldKey], size
    ) -> Mapping[FieldKey, np.ndarray]:
        # This function has an interesting history.  It previously was mandate
        # to be defined by all of the subclasses.  But, to avoid having to
        # rewrite a whole bunch of IO handlers all at once, and to allow a
        # better abstraction for grid-based frontends, we're now defining it in
        # the base class.
        rv = {}
        nodal_fields = []
        for field in fields:
            finfo = self.ds.field_info[field]
            nodal_flag = finfo.nodal_flag
            if np.any(nodal_flag):
                num_nodes = 2 ** sum(nodal_flag)
                rv[field] = np.empty((size, num_nodes), dtype="=f8")
                nodal_fields.append(field)
            else:
                rv[field] = np.empty(size, dtype="=f8")
        ind = {field: 0 for field in fields}
        for field, obj, data in self.io_iter(chunks, fields):
            if data is None:
                continue
            if isinstance(selector, GridSelector) and field not in nodal_fields:
                ind[field] += data.size
                rv[field] = data.copy()
            else:
                ind[field] += obj.select(selector, data, rv[field], ind[field])
        return rv

    def io_iter(self, chunks, fields: list[FieldKey]):
        raise NotImplementedError(
            "subclassing Dataset.io_iter this is required in order to use the default "
            "implementation of Dataset._read_fluid_selection. "
            "Custom implementations of the latter may not rely on this method."
        )

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        tr = self._read_data_set(grid, field)[tuple(sl)]
        if tr.dtype == "float32":
            tr = tr.astype("float64")
        return tr

    def _read_field_names(self, grid):
        pass

    @property
    def _read_exception(self):
        return None

    def _read_chunk_data(self, chunk, fields):
        return {}

    def _read_particle_coords(
        self, chunks, ptf: defaultdict[str, list[str]]
    ) -> Iterator[ParticleCoordinateTuple]:
        # An iterator that yields particle coordinates for each chunk by particle
        # type. Must be implemented by each frontend. Must yield a tuple of
        # (particle type, xyz, hsml) by chunk. If the frontend does not have
        # a smoothing length, yield (particle type, xyz, 0.0)
        raise NotImplementedError

    def _read_particle_data_file(self, data_file, ptf, selector=None):
        # each frontend needs to implement this: read from a data_file object
        # and return a dict of fields for that data_file
        raise NotImplementedError

    def _read_particle_selection(
        self, chunks, selector, fields: list[FieldKey]
    ) -> dict[FieldKey, np.ndarray]:
        data: dict[FieldKey, list[np.ndarray]] = {}

        # Initialize containers for tracking particle, field information
        # ptf (particle field types) maps particle type to list of on-disk fields to read
        # field_maps stores fields, accounting for field unions
        ptf: defaultdict[str, list[str]] = defaultdict(list)
        field_maps: defaultdict[FieldKey, list[FieldKey]] = defaultdict(list)

        # We first need a set of masks for each particle type
        chunks = list(chunks)
        unions = self.ds.particle_unions
        # What we need is a mapping from particle types to return types
        for field in fields:
            ftype, fname = field
            # We should add a check for p.fparticle_unions or something here
            if ftype in unions:
                for pt in unions[ftype]:
                    ptf[pt].append(fname)
                    field_maps[pt, fname].append(field)
            else:
                ptf[ftype].append(fname)
                field_maps[field].append(field)
            data[field] = []

        # Now we read.
        for field_r, vals in self._read_particle_fields(chunks, ptf, selector):
            # Note that we now need to check the mappings
            for field_f in field_maps[field_r]:
                data[field_f].append(vals)

        rv: dict[FieldKey, np.ndarray] = {}  # the return dictionary
        fields = list(data.keys())
        for field_f in fields:
            # We need to ensure the arrays have the right shape if there are no
            # particles in them.
            total = sum(_.size for _ in data[field_f])
            if total > 0:
                vals = data.pop(field_f)
                # note: numpy.concatenate has a dtype argument that would avoid
                # a copy using .astype(...), available in numpy>=1.20
                rv[field_f] = np.concatenate(vals, axis=0).astype("float64")
            else:
                shape = [0]
                if field_f[1] in self._vector_fields:
                    shape.append(self._vector_fields[field_f[1]])
                elif field_f[1] in self._array_fields:
                    shape.append(self._array_fields[field_f[1]])
                rv[field_f] = np.empty(shape, dtype="float64")
        return rv

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        for data_file in self._sorted_chunk_iterator(chunks):
            data_file_data = self._read_particle_data_file(data_file, ptf, selector)
            # temporary trickery so it's still an iterator, need to adjust
            # the io_handler.BaseIOHandler.read_particle_selection() method
            # to not use an iterator.
            yield from data_file_data.items()

    @staticmethod
    def _get_data_files(chunks):
        chunks = list(chunks)
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        return data_files

    def _sorted_chunk_iterator(self, chunks):
        data_files = self._get_data_files(chunks)
        yield from sorted(data_files, key=lambda x: (x.filename, x.start))


# As a note: we don't *actually* want this to be how it is forever.  There's no
# reason we need to have the fluid and particle IO handlers separated.  But,
# for keeping track of which frontend is which, this is a useful abstraction.
class BaseParticleIOHandler(BaseIOHandler):
    pass


class IOHandlerExtracted(BaseIOHandler):
    _dataset_type = "extracted"

    def _read_data_set(self, grid, field):
        return grid.base_grid[field] / grid.base_grid.convert(field)

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        return grid.base_grid[field][tuple(sl)] / grid.base_grid.convert(field)



import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
)
from yt.funcs import mylog
from yt.utilities.lib.mesh_utilities import fill_fcoords, fill_fwidths


class UnstructuredMesh(YTSelectionContainer):
    # This is a base class, not meant to be used directly.
    _spatial = False
    _connectivity_length = -1
    _type_name = "unstructured_mesh"
    _skip_add = True
    _index_offset = 0
    _con_args = ("mesh_id", "filename", "connectivity_indices", "connectivity_coords")

    def __init__(
        self, mesh_id, filename, connectivity_indices, connectivity_coords, index
    ):
        super().__init__(index.dataset, None)
        self.filename = filename
        self.mesh_id = mesh_id
        # This is where we set up the connectivity information
        self.connectivity_indices = connectivity_indices
        if connectivity_indices.shape[1] != self._connectivity_length:
            if self._connectivity_length == -1:
                self._connectivity_length = connectivity_indices.shape[1]
            else:
                raise RuntimeError
        self.connectivity_coords = connectivity_coords
        self.ds = index.dataset
        self._index = index
        self._last_mask = None
        self._last_count = -1
        self._last_selector_id = None

    def _check_consistency(self):
        if self.connectivity_indices.shape[1] != self._connectivity_length:
            raise RuntimeError

        for gi in range(self.connectivity_indices.shape[0]):
            ind = self.connectivity_indices[gi, :] - self._index_offset
            coords = self.connectivity_coords[ind, :]
            for i in range(3):
                assert np.unique(coords[:, i]).size == 2
        mylog.debug("Connectivity is consistent.")

    def __repr__(self):
        return "UnstructuredMesh_%04i" % (self.mesh_id)

    def get_global_startindex(self):
        """
        Return the integer starting index for each dimension at the current
        level.

        """
        raise NotImplementedError

    def convert(self, datatype):
        """
        This will attempt to convert a given unit to cgs from code units. It
        either returns the multiplicative factor or throws a KeyError.

        """
        return self.ds[datatype]

    @property
    def shape(self):
        raise NotImplementedError

    def _generate_container_field(self, field):
        raise NotImplementedError

    def select_fcoords(self, dobj=None):
        # This computes centroids!
        mask = self._get_selector_mask(dobj.selector)
        if mask is None:
            return np.empty((0, 3), dtype="float64")
        centers = fill_fcoords(
            self.connectivity_coords, self.connectivity_indices, self._index_offset
        )
        return centers[mask, :]

    def select_fwidth(self, dobj):
        raise NotImplementedError

    def select_icoords(self, dobj):
        raise NotImplementedError

    def select_ires(self, dobj):
        raise NotImplementedError

    def select_tcoords(self, dobj):
        raise NotImplementedError

    def deposit(self, positions, fields=None, method=None, kernel_name="cubic"):
        raise NotImplementedError

    def select_blocks(self, selector):
        mask = self._get_selector_mask(selector)
        yield self, mask

    def select(self, selector, source, dest, offset):
        mask = self._get_selector_mask(selector)
        count = self.count(selector)
        if count == 0:
            return 0
        dest[offset : offset + count] = source[mask, ...]
        return count

    def count(self, selector):
        mask = self._get_selector_mask(selector)
        if mask is None:
            return 0
        return self._last_count

    def count_particles(self, selector, x, y, z):
        # We don't cache the selector results
        count = selector.count_points(x, y, z, 0.0)
        return count

    def select_particles(self, selector, x, y, z):
        mask = selector.select_points(x, y, z, 0.0)
        return mask

    def _get_selector_mask(self, selector):
        if hash(selector) == self._last_selector_id:
            mask = self._last_mask
        else:
            self._last_mask = mask = selector.fill_mesh_cell_mask(self)
            self._last_selector_id = hash(selector)
            if mask is None:
                self._last_count = 0
            else:
                self._last_count = mask.sum()
        return mask

    def select_fcoords_vertex(self, dobj=None):
        mask = self._get_selector_mask(dobj.selector)
        if mask is None:
            return np.empty((0, self._connectivity_length, 3), dtype="float64")
        vertices = self.connectivity_coords[self.connectivity_indices - 1]
        return vertices[mask, :, :]


class SemiStructuredMesh(UnstructuredMesh):
    _connectivity_length = 8
    _type_name = "semi_structured_mesh"
    _container_fields = ("dx", "dy", "dz")

    def __repr__(self):
        return "SemiStructuredMesh_%04i" % (self.mesh_id)

    def _generate_container_field(self, field):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        if field == "dx":
            return self._current_chunk.fwidth[:, 0]
        elif field == "dy":
            return self._current_chunk.fwidth[:, 1]
        elif field == "dz":
            return self._current_chunk.fwidth[:, 2]

    def select_fwidth(self, dobj):
        mask = self._get_selector_mask(dobj.selector)
        if mask is None:
            return np.empty((0, 3), dtype="float64")
        widths = fill_fwidths(
            self.connectivity_coords, self.connectivity_indices, self._index_offset
        )
        return widths[mask, :]

    def select_ires(self, dobj):
        ind = np.zeros(self.connectivity_indices.shape[0])
        mask = self._get_selector_mask(dobj.selector)
        if mask is None:
            return np.empty(0, dtype="int32")
        return ind[mask]

    def select_tcoords(self, dobj):
        mask = self._get_selector_mask(dobj.selector)
        if mask is None:
            return np.empty(0, dtype="float64")
        dt, t = dobj.selector.get_dt_mesh(self, mask.sum(), self._index_offset)
        return dt, t

    def _get_selector_mask(self, selector):
        if hash(selector) == self._last_selector_id:
            mask = self._last_mask
        else:
            self._last_mask = mask = selector.fill_mesh_cell_mask(self)
            self._last_selector_id = hash(selector)
            if mask is None:
                self._last_count = 0
            else:
                self._last_count = mask.sum()
        return mask

    def select(self, selector, source, dest, offset):
        mask = self._get_selector_mask(selector)
        count = self.count(selector)
        if count == 0:
            return 0
        # Note: this likely will not work with vector fields.
        dest[offset : offset + count] = source.flat[mask]
        return count



from collections import defaultdict

import numpy as np

from yt.frontends.gadget_fof.io import IOHandlerGadgetFOFHaloHDF5
from yt.funcs import parse_h5_attr
from yt.units._numpy_wrapper_functions import uvstack
from yt.utilities.io_handler import BaseParticleIOHandler
from yt.utilities.on_demand_imports import _h5py as h5py


class IOHandlerYTHaloCatalog(BaseParticleIOHandler):
    _dataset_type = "ythalocatalog"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == "halos"
        ptype = "halos"
        pn = "particle_position_%s"
        for data_file in self._sorted_chunk_iterator(chunks):
            with h5py.File(data_file.filename, mode="r") as f:
                units = parse_h5_attr(f[pn % "x"], "units")
                pos = data_file._get_particle_positions(ptype, f=f)
                x, y, z = (self.ds.arr(pos[:, i], units) for i in range(3))
                yield "halos", (x, y, z), 0.0

    def _yield_coordinates(self, data_file):
        pn = "particle_position_%s"
        with h5py.File(data_file.filename, mode="r") as f:
            units = parse_h5_attr(f[pn % "x"], "units")
            x, y, z = (
                self.ds.arr(f[pn % ax][()].astype("float64"), units) for ax in "xyz"
            )
            pos = uvstack([x, y, z]).T
            pos.convert_to_units("code_length")
            yield "halos", pos

    def _read_particle_fields(self, chunks, ptf, selector):
        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == "halos"
        pn = "particle_position_%s"
        for data_file in self._sorted_chunk_iterator(chunks):
            si, ei = data_file.start, data_file.end
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype, field_list in sorted(ptf.items()):
                    units = parse_h5_attr(f[pn % "x"], "units")
                    pos = data_file._get_particle_positions(ptype, f=f)
                    x, y, z = (self.ds.arr(pos[:, i], units) for i in range(3))
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    if mask is None:
                        continue
                    for field in field_list:
                        data = f[field][si:ei][mask].astype("float64")
                        yield (ptype, field), data

    def _count_particles(self, data_file):
        si, ei = data_file.start, data_file.end
        nhalos = data_file.header["num_halos"]
        if None not in (si, ei):
            nhalos = np.clip(nhalos - si, 0, ei - si)
        return {"halos": nhalos}

    def _identify_fields(self, data_file):
        with h5py.File(data_file.filename, mode="r") as f:
            fields = [
                ("halos", field) for field in f if not isinstance(f[field], h5py.Group)
            ]
            units = {("halos", field): parse_h5_attr(f[field], "units") for field in f}
        return fields, units


class HaloDatasetIOHandler:
    """
    Base class for io handlers to load halo member particles.
    """

    def _read_particle_coords(self, chunks, ptf):
        pass

    def _read_particle_fields(self, dobj, ptf):
        # separate member particle fields from scalar fields
        scalar_fields = defaultdict(list)
        member_fields = defaultdict(list)
        for ptype, field_list in sorted(ptf.items()):
            for field in field_list:
                if (ptype, field) in self.ds.scalar_field_list:
                    scalar_fields[ptype].append(field)
                else:
                    member_fields[ptype].append(field)

        all_data = self._read_scalar_fields(dobj, scalar_fields)
        all_data.update(self._read_member_fields(dobj, member_fields))

        for field, field_data in all_data.items():
            yield field, field_data

    # This will be refactored.
    _read_particle_selection = IOHandlerGadgetFOFHaloHDF5._read_particle_selection


# ignoring type in this mixing to circumvent this error from mypy
# Definition of "_read_particle_fields" in base class "HaloDatasetIOHandler"
# is incompatible with definition in base class "IOHandlerYTHaloCatalog"
#
# it may not be possible to refactor out of this situation without breaking downstream
class IOHandlerYTHalo(HaloDatasetIOHandler, IOHandlerYTHaloCatalog):  # type: ignore
    _dataset_type = "ythalo"

    def _identify_fields(self, data_file):
        with h5py.File(data_file.filename, mode="r") as f:
            scalar_fields = [
                ("halos", field) for field in f if not isinstance(f[field], h5py.Group)
            ]
            units = {("halos", field): parse_h5_attr(f[field], "units") for field in f}
            if "particles" in f:
                id_fields = [("halos", field) for field in f["particles"]]
            else:
                id_fields = []

        return scalar_fields + id_fields, scalar_fields, id_fields, units

    def _read_member_fields(self, dobj, member_fields):
        all_data = defaultdict(lambda: np.empty(dobj.particle_number, dtype=np.float64))
        if not member_fields:
            return all_data
        field_start = 0
        for i, data_file in enumerate(dobj.field_data_files):
            start_index = dobj.field_data_start[i]
            end_index = dobj.field_data_end[i]
            pcount = end_index - start_index
            if pcount == 0:
                continue
            field_end = field_start + end_index - start_index
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype, field_list in sorted(member_fields.items()):
                    for field in field_list:
                        field_data = all_data[ptype, field]
                        my_data = f["particles"][field][start_index:end_index].astype(
                            "float64"
                        )
                        field_data[field_start:field_end] = my_data
            field_start = field_end
        return all_data

    def _read_scalar_fields(self, dobj, scalar_fields):
        all_data = {}
        if not scalar_fields:
            return all_data
        with h5py.File(dobj.scalar_data_file.filename, mode="r") as f:
            for ptype, field_list in sorted(scalar_fields.items()):
                for field in field_list:
                    data = np.array([f[field][dobj.scalar_index]]).astype("float64")
                    all_data[ptype, field] = data
        return all_data



import numpy as np
from numpy.testing import assert_array_equal, assert_equal

import yt.utilities.lib.bitarray as ba


def test_inout_bitarray():
    # Check that we can do it for bitarrays that are funny-shaped
    rng = np.random.default_rng()
    for i in range(7):
        # Check we can feed in an array
        arr_in = rng.random(32**3 + i) > 0.5
        b = ba.bitarray(arr=arr_in)
        if i > 0:
            assert_equal(b.ibuf.size, (32**3) / 8.0 + 1)
        arr_out = b.as_bool_array()
        assert_equal(arr_in, arr_out)

        # Let's check we can do it without feeding it at first
        b = ba.bitarray(size=arr_in.size)
        b.set_from_array(arr_in)
        arr_out = b.as_bool_array()
        assert_equal(arr_in, arr_out)

    # Try a big array
    arr_in = rng.random(32**3 + i) > 0.5
    b = ba.bitarray(arr=arr_in)
    arr_out = b.as_bool_array()
    assert_equal(arr_in, arr_out)
    assert_equal(b.count(), arr_in.sum())

    # Let's check we can do something interesting.
    arr_in1 = rng.random(32**3) > 0.5
    arr_in2 = rng.random(32**3) > 0.5
    b1 = ba.bitarray(arr=arr_in1)
    b2 = ba.bitarray(arr=arr_in2)
    b3 = ba.bitarray(arr=(arr_in1 & arr_in2))
    assert_equal((b1.ibuf & b2.ibuf), b3.ibuf)
    assert_equal(b1.count(), arr_in1.sum())
    assert_equal(b2.count(), arr_in2.sum())
    # Let's check the logical and operation
    b4 = b1.logical_and(b2)
    assert_equal(b4.count(), b3.count())
    assert_array_equal(b4.as_bool_array(), b3.as_bool_array())

    b5 = b1 & b2
    assert_equal(b5.count(), b3.count())
    assert_array_equal(b5.as_bool_array(), b3.as_bool_array())

    b1 &= b2
    assert_equal(b1.count(), b4.count())
    assert_array_equal(b1.as_bool_array(), b4.as_bool_array())

    # Repeat this, but with the logical or operators
    b1 = ba.bitarray(arr=arr_in1)
    b2 = ba.bitarray(arr=arr_in2)
    b3 = ba.bitarray(arr=(arr_in1 | arr_in2))
    assert_equal((b1.ibuf | b2.ibuf), b3.ibuf)
    assert_equal(b1.count(), arr_in1.sum())
    assert_equal(b2.count(), arr_in2.sum())
    # Let's check the logical and operation
    b4 = b1.logical_or(b2)
    assert_equal(b4.count(), b3.count())
    assert_array_equal(b4.as_bool_array(), b3.as_bool_array())

    b5 = b1 | b2
    assert_equal(b5.count(), b3.count())
    assert_array_equal(b5.as_bool_array(), b3.as_bool_array())

    b1 |= b2
    assert_equal(b1.count(), b4.count())
    assert_array_equal(b1.as_bool_array(), b4.as_bool_array())

    # Repeat this, but with the logical xor operators
    b1 = ba.bitarray(arr=arr_in1)
    b2 = ba.bitarray(arr=arr_in2)
    b3 = ba.bitarray(arr=(arr_in1 ^ arr_in2))
    assert_equal((b1.ibuf ^ b2.ibuf), b3.ibuf)
    assert_equal(b1.count(), arr_in1.sum())
    assert_equal(b2.count(), arr_in2.sum())
    # Let's check the logical and operation
    b4 = b1.logical_xor(b2)
    assert_equal(b4.count(), b3.count())
    assert_array_equal(b4.as_bool_array(), b3.as_bool_array())

    b5 = b1 ^ b2
    assert_equal(b5.count(), b3.count())
    assert_array_equal(b5.as_bool_array(), b3.as_bool_array())

    b1 ^= b2
    assert_equal(b1.count(), b4.count())
    assert_array_equal(b1.as_bool_array(), b4.as_bool_array())

    b = ba.bitarray(10)
    for i in range(10):
        b.set_value(i, 2)  # 2 should evaluate to True
        arr = b.as_bool_array()
        assert_equal(arr[: i + 1].all(), True)
        assert_equal(arr[i + 1 :].any(), False)
    for i in range(10):
        b.set_value(i, 0)
    arr = b.as_bool_array()
    assert_equal(arr.any(), False)
    b.set_value(7, 1)
    arr = b.as_bool_array()
    assert_array_equal(arr, [0, 0, 0, 0, 0, 0, 0, 1, 0, 0])
    b.set_value(2, 1)
    arr = b.as_bool_array()
    assert_array_equal(arr, [0, 0, 1, 0, 0, 0, 0, 1, 0, 0])


def test_set_range():
    b = ba.bitarray(127)
    # Test once where we're in the middle of start and end bits
    b.set_range(4, 65, 1)
    comparison_array = np.zeros(127, dtype="uint8")
    comparison_array[4:65] = 1
    arr = b.as_bool_array().astype("uint8")
    assert_array_equal(arr, comparison_array)
    assert_equal(b.count(), comparison_array.sum())

    # Test when we start and stop in the same byte
    b = ba.bitarray(127)
    b.set_range(4, 6, 1)
    comparison_array = np.zeros(127, dtype="uint8")
    comparison_array[4:6] = 1
    arr = b.as_bool_array().astype("uint8")
    assert_array_equal(arr, comparison_array)
    assert_equal(b.count(), comparison_array.sum())

    # Test now where we're in the middle of start
    b = ba.bitarray(64)
    b.set_range(33, 36, 1)
    comparison_array = np.zeros(64, dtype="uint8")
    comparison_array[33:36] = 1
    arr = b.as_bool_array().astype("uint8")
    assert_array_equal(arr, comparison_array)
    assert_equal(b.count(), comparison_array.sum())

    # Now we test when we end on a byte edge, but we have 65 entries
    b = ba.bitarray(65)
    b.set_range(32, 64, 1)
    comparison_array = np.zeros(65, dtype="uint8")
    comparison_array[32:64] = 1
    arr = b.as_bool_array().astype("uint8")
    assert_array_equal(arr, comparison_array)
    assert_equal(b.count(), comparison_array.sum())

    # Let's do the inverse
    b = ba.bitarray(127)
    b.set_range(0, 127, 1)
    assert_equal(b.as_bool_array().all(), True)
    b.set_range(0, 127, 0)
    assert_equal(b.as_bool_array().any(), False)
    b.set_range(3, 9, 1)
    comparison_array = np.zeros(127, dtype="uint8")
    comparison_array[3:9] = 1
    arr = b.as_bool_array().astype("uint8")
    assert_array_equal(arr, comparison_array)
    assert_equal(b.count(), comparison_array.sum())

    # Now let's overlay some zeros
    b.set_range(7, 10, 0)
    comparison_array[7:10] = 0
    arr = b.as_bool_array().astype("uint8")
    assert_array_equal(arr, comparison_array)



from itertools import groupby

import numpy as np

from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.logger import ytLogger as mylog


# http://stackoverflow.com/questions/2361945/detecting-consecutive-integers-in-a-list
def grid_sequences(grids):
    g_iter = sorted(grids, key=lambda g: g.id)
    for _, g in groupby(enumerate(g_iter), lambda i_x1: i_x1[0] - i_x1[1].id):
        seq = [v[1] for v in g]
        yield seq


ii = [0, 1, 0, 1, 0, 1, 0, 1]
jj = [0, 0, 1, 1, 0, 0, 1, 1]
kk = [0, 0, 0, 0, 1, 1, 1, 1]


class IOHandlerParthenon(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "parthenon"

    def __init__(self, ds):
        super().__init__(ds)
        self._handle = ds._handle

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        f = self._handle
        rv = {}
        for field in fields:
            # Always use *native* 64-bit float.
            rv[field] = np.empty(size, dtype="=f8")
        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s blocks",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )
        last_dname = None
        for field in fields:
            ftype, fname = field
            dname, fdi = self.ds._field_map[fname]
            if dname != last_dname:
                ds = f[f"/{dname}"]
            ind = 0
            for chunk in chunks:
                for gs in grid_sequences(chunk.objs):
                    start = gs[0].id - gs[0]._id_offset
                    end = gs[-1].id - gs[-1]._id_offset + 1
                    if len(ds.shape) == 4:
                        data = ds[start:end, :, :, :].transpose()
                    else:
                        data = ds[start:end, fdi, :, :, :].transpose()
                    for i, g in enumerate(gs):
                        ind += g.select(selector, data[..., i], rv[field], ind)
            last_dname = dname
        return rv

    def _read_chunk_data(self, chunk, fields):
        f = self._handle
        rv = {}
        for g in chunk.objs:
            rv[g.id] = {}
        if len(fields) == 0:
            return rv
        for field in fields:
            ftype, fname = field
            dname, fdi = self.ds._field_map[fname]
            ds = f[f"/{dname}"]
            for gs in grid_sequences(chunk.objs):
                start = gs[0].id - gs[0]._id_offset
                end = gs[-1].id - gs[-1]._id_offset + 1
                if len(ds.shape) == 4:
                    data = ds[start:end, :, :, :].transpose()
                else:
                    data = ds[start:end, fdi, :, :, :].transpose()
                for i, g in enumerate(gs):
                    rv[g.id][field] = np.asarray(data[..., i], "=f8")
        return rv



import os
import shutil
import tempfile
from unittest import TestCase

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from yt.testing import fake_random_ds


def setup_module():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_flux_calculation():
    ds = fake_random_ds(64, nprocs=4)
    dd = ds.all_data()
    surf = ds.surface(dd, ("index", "x"), 0.51)
    assert_equal(surf["index", "x"], 0.51)
    flux = surf.calculate_flux(
        ("index", "ones"), ("index", "zeros"), ("index", "zeros"), ("index", "ones")
    )
    assert_almost_equal(flux.value, 1.0, 12)
    assert_equal(str(flux.units), "cm**2")
    flux2 = surf.calculate_flux(
        ("index", "ones"), ("index", "zeros"), ("index", "zeros")
    )
    assert_almost_equal(flux2.value, 1.0, 12)
    assert_equal(str(flux2.units), "cm**2")


def test_sampling():
    ds = fake_random_ds(64, nprocs=4)
    dd = ds.all_data()
    for i, ax in enumerate([("index", "x"), ("index", "y"), ("index", "z")]):
        surf = ds.surface(dd, ax, 0.51)
        surf.get_data(ax, sample_type="vertex")
        assert_equal(surf.vertex_samples[ax], surf.vertices[i, :])
        assert_equal(str(surf.vertices.units), "code_length")
        dens = surf["gas", "density"]
        vert_shape = surf.vertices.shape
        assert_equal(dens.shape[0], vert_shape[1] // vert_shape[0])
        assert_equal(str(dens.units), "g/cm**3")


class ExporterTests(TestCase):
    def setUp(self):
        self.curdir = os.getcwd()
        self.tmpdir = tempfile.mkdtemp()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)

    def test_export_ply(self):
        ds = fake_random_ds(64, nprocs=4)
        dd = ds.all_data()
        surf = ds.surface(dd, ("index", "x"), 0.51)
        surf.export_ply("my_ply.ply", bounds=[(0, 1), (0, 1), (0, 1)])
        assert os.path.exists("my_ply.ply")
        surf.export_ply(
            "my_ply2.ply",
            bounds=[(0, 1), (0, 1), (0, 1)],
            sample_type="vertex",
            color_field=("gas", "density"),
        )
        assert os.path.exists("my_ply2.ply")

    def test_export_obj(self):
        ds = fake_random_ds(
            16,
            nprocs=4,
            particles=16**3,
            fields=("density", "temperature"),
            units=("g/cm**3", "K"),
        )
        sp = ds.sphere("max", (1.0, "cm"))
        surf = ds.surface(sp, ("gas", "density"), 0.5)
        surf.export_obj("my_galaxy", transparency=1.0, dist_fac=1.0)
        assert os.path.exists("my_galaxy.obj")
        assert os.path.exists("my_galaxy.mtl")

        mi, ma = sp.quantities.extrema(("gas", "temperature"))
        rhos = [0.5, 0.25]
        trans = [0.5, 1.0]
        for i, r in enumerate(rhos):
            basename = "my_galaxy_color"
            surf = ds.surface(sp, ("gas", "density"), r)
            surf.export_obj(
                basename,
                transparency=trans[i],
                color_field=("gas", "temperature"),
                dist_fac=1.0,
                plot_index=i,
                color_field_max=ma,
                color_field_min=mi,
            )

            assert os.path.exists(f"{basename}.obj")
            assert os.path.exists(f"{basename}.mtl")

        def _Emissivity(field, data):
            return (
                data["gas", "density"]
                * data["gas", "density"]
                * np.sqrt(data["gas", "temperature"])
            )

        ds.add_field(
            ("gas", "emissivity"),
            sampling_type="cell",
            function=_Emissivity,
            units=r"g**2*sqrt(K)/cm**6",
        )
        for i, r in enumerate(rhos):
            basename = "my_galaxy_emis"
            surf = ds.surface(sp, ("gas", "density"), r)
            surf.export_obj(
                basename,
                transparency=trans[i],
                color_field=("gas", "temperature"),
                emit_field=("gas", "emissivity"),
                dist_fac=1.0,
                plot_index=i,
            )

            basename = "my_galaxy_emis"
            assert os.path.exists(f"{basename}.obj")
            assert os.path.exists(f"{basename}.mtl")


def test_correct_output_unit_fake_ds():
    # see issue #1368
    ds = fake_random_ds(64, nprocs=4, particles=16**3)
    x = y = z = 0.5
    sp1 = ds.sphere((x, y, z), (300, "kpc"))
    Nmax = sp1.max(("gas", "density"))
    sur = ds.surface(sp1, ("gas", "density"), 0.5 * Nmax)
    sur["index", "x"][0]


def test_radius_surface():
    # see #1407
    ds = fake_random_ds(64, nprocs=4, particles=16**3, length_unit=10.0)
    reg = ds.all_data()
    sp = ds.sphere(ds.domain_center, (0.5, "code_length"))
    for obj in [reg, sp]:
        for rad in [0.05, 0.1, 0.4]:
            surface = ds.surface(obj, ("index", "radius"), (rad, "code_length"))
            assert_almost_equal(surface.surface_area.v, 4 * np.pi * rad**2, decimal=2)
            verts = surface.vertices
            for i in range(3):
                assert_almost_equal(verts[i, :].min().v, 0.5 - rad, decimal=2)
                assert_almost_equal(verts[i, :].max().v, 0.5 + rad, decimal=2)



class ValidationException(Exception):
    pass


class NeedsGridType(ValidationException):
    def __init__(self, ghost_zones=0, fields=None):
        self.ghost_zones = ghost_zones
        self.fields = fields

    def __str__(self):
        s = "s" if self.ghost_zones != 1 else ""
        return f"fields {self.fields} require {self.ghost_zones} ghost zone{s}."


class NeedsOriginalGrid(NeedsGridType):
    def __init__(self):
        self.ghost_zones = 0


class NeedsDataField(ValidationException):
    def __init__(self, missing_fields):
        self.missing_fields = missing_fields

    def __str__(self):
        return f"({self.missing_fields})"


class NeedsProperty(ValidationException):
    def __init__(self, missing_properties):
        self.missing_properties = missing_properties

    def __str__(self):
        return f"({self.missing_properties})"


class NeedsParameter(ValidationException):
    def __init__(self, missing_parameters):
        self.missing_parameters = missing_parameters

    def __str__(self):
        return f"({self.missing_parameters})"


class NeedsConfiguration(ValidationException):
    def __init__(self, parameter, value):
        self.parameter = parameter
        self.value = value

    def __str__(self):
        return f"(Needs {self.parameter} = {self.value})"


class FieldUnitsError(Exception):
    pass



from io import BytesIO

import PIL
from PIL import Image
from PIL.PngImagePlugin import PngInfo

from .._version import __version__ as yt_version


def call_png_write_png(buffer, fileobj, dpi):
    metadata = PngInfo()
    metadata.add_text("Software", f"PIL-{PIL.__version__}|yt-{yt_version}")
    Image.fromarray(buffer).save(
        fileobj, dpi=(dpi, dpi), format="png", pnginfo=metadata
    )


def write_png(buffer, filename, dpi=100):
    with open(filename, "wb") as fileobj:
        call_png_write_png(buffer, fileobj, dpi)


def write_png_to_string(buffer, dpi=100):
    fileobj = BytesIO()
    call_png_write_png(buffer, fileobj, dpi)
    png_str = fileobj.getvalue()
    fileobj.close()
    return png_str
